import os.path
import shutil
import sys
import tempfile
import textwrap

from egginst.tests.common import create_venv, mkdtemp
from egginst.utils import (atomic_file, get_executable, parse_assignments,
                           samefile)
from egginst.vendor.six import StringIO
from egginst.vendor.six.moves import unittest
from enstaller.errors import InvalidFormat


class TestParseAssignments(unittest.TestCase):
    def test_parse_simple(self):
        r_data = {"IndexedRepos": ["http://acme.com/{SUBDIR}"],
                  "webservice_entry_point": "http://acme.com/eggs/{PLATFORM}/"}

        s = textwrap.dedent("""\
        IndexedRepos = [
            "http://acme.com/{SUBDIR}",
        ]
        webservice_entry_point = "http://acme.com/eggs/{PLATFORM}/"
        """)

        data = parse_assignments(StringIO(s))
        self.assertEqual(data, r_data)

    def test_parse_simple_invalid_file(self):
        with self.assertRaises(InvalidFormat):
            parse_assignments(StringIO("EPD_auth += 2"))

        with self.assertRaises(InvalidFormat):
            parse_assignments(StringIO("1 + 2"))


class TestAtomicFile(unittest.TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def test_simple(self):
        # Given
        r_content = b"some data"
        path = os.path.join(self.tempdir, "some_data.bin")

        # When
        with atomic_file(path) as fp:
            fp.write(r_content)

        # Then
        self.assertTrue(os.path.exists(path))
        with open(path, "rb") as fp:
            content = fp.read()
        self.assertEqual(content, r_content)

    def test_failure(self):
        # Given
        r_content = b"some data"
        path = os.path.join(self.tempdir, "some_data.bin")

        # When
        try:
            with atomic_file(path) as fp:
                fp.write(r_content[:2])
                raise KeyboardInterrupt()
        except BaseException:
            pass

        # Then
        self.assertFalse(os.path.exists(path))

    def test_abort(self):
        # Given
        r_content = b"some data"
        path = os.path.join(self.tempdir, "some_data.bin")

        # When
        with atomic_file(path) as fp:
            fp.write(r_content[:2])
            temp_name = fp._name
            fp.abort()

        # Then
        self.assertFalse(os.path.exists(path))
        self.assertFalse(os.path.exists(temp_name))

    def test_exists(self):
        # Given
        r_content = b"some data"
        path = os.path.join(self.tempdir, "some_data.bin")
        with open(path, "wb") as fp:
            fp.write(b"nono")

        # When
        with atomic_file(path) as fp:
            fp.write(r_content)

        # Then
        self.assertTrue(os.path.exists(path))
        with open(path, "rb") as fp:
            content = fp.read()
        self.assertEqual(content, r_content)

    def test_exists_with_failure(self):
        # Given
        r_content = b"nono"
        path = os.path.join(self.tempdir, "some_data.bin")
        with open(path, "wb") as fp:
            fp.write(r_content)

        # When
        try:
            with atomic_file(path) as fp:
                fp.write(b"some data")
                raise ValueError("some random failure")
        except BaseException:
            pass

        # Then
        self.assertTrue(os.path.exists(path))
        with open(path, "rb") as fp:
            content = fp.read()
        self.assertEqual(content, r_content)


class TestSameFile(unittest.TestCase):
    def setUp(self):
        self.prefix = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.prefix)

    def test_simple_same_file(self):
        # Given
        os.makedirs(os.path.join(self.prefix, "foo"))
        left = os.path.join(self.prefix, "bar")
        right = os.path.join(self.prefix, "foo", os.pardir, "bar")

        with open(left, "w") as fp:
            fp.write("")

        # When/Then
        self.assertTrue(samefile(left, right))

    def test_simple_not_same_file(self):
        # Given
        left = os.path.join(self.prefix, "bar")
        right = os.path.join(self.prefix, "foo")

        for path in left, right:
            with open(path, "w") as fp:
                fp.write("")

        # When/Then
        self.assertFalse(samefile(left, right))


class TestGetExecutable(unittest.TestCase):
    def test_simple(self):
        # Given
        with mkdtemp() as prefix:
            if sys.platform == "win32":
                # python.exe is in scripts because we use virtualenv
                r_executable = os.path.join(prefix, "Scripts", "python.exe")
            else:
                r_executable = os.path.join(prefix, "bin", "python")

            create_venv(prefix)

            # When
            executable = get_executable(prefix)

        # Then
        self.assertEqual(executable, r_executable)



import os.path
import tempfile
import textwrap

from egginst.vendor.six.moves import unittest
from enstaller.eggcollect import info_from_metadir


class TestInfoFromMetaDir(unittest.TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()

    def tearDown(self):
        self.tempdir = tempfile.mkdtemp()

    def test_simple(self):
        # Given
        data = textwrap.dedent("""{
            "arch": "amd64",
            "build": 1,
            "ctime": "Thu Apr 24 15:41:24 2014",
            "hook": false,
            "key": "VTK-5.10.1-1.egg",
            "name": "vtk",
            "osdist": "RedHat_5",
            "packages": [],
            "platform": "linux2",
            "python": "2.7",
            "type": "egg",
            "version": "5.10.1"
        }""")
        r_info = {
            "arch": "amd64",
            "build": 1,
            "ctime": "Thu Apr 24 15:41:24 2014",
            "hook": False,
            "installed": True,
            "key": "VTK-5.10.1-1.egg",
            "meta_dir": self.tempdir,
            "name": "vtk",
            "osdist": "RedHat_5",
            "packages": [],
            "platform": "linux2",
            "python": "2.7",
            "type": "egg",
            "version": "5.10.1"
        }
        info_json = os.path.join(self.tempdir, "_info.json")
        with open(info_json, "w") as fp:
            fp.write(data)

        # When
        info = info_from_metadir(self.tempdir)

        # Then
        self.assertEqual(info, r_info)



import sys

from egginst.console import ProgressBar, get_terminal_size
from egginst.console.simple import _MAX_SPEED_LABEL_DISPLAY
from egginst.utils import human_bytes


def dummy_progress_bar_factory(*a, **kw):
    return _DummyProgressBar()


def _compute_optimal_first_line(message, filename):
    term_width = get_terminal_size()[0]
    # -------------------------------------------- term width
    # first_line_left first_line_right speed_label
    # The '1' are for spaces
    first_line_length = term_width - 1 - _MAX_SPEED_LABEL_DISPLAY
    first_line_right = 20
    first_line_left = first_line_length - first_line_right - 1

    first_line_template = "%%-%ss %%%ss" % (first_line_left, first_line_right)
    if len(filename) >= first_line_left:
        ellipsis = "... "
        display_filename = (filename[:first_line_left - len(ellipsis)] +
                            ellipsis)
    else:
        display_filename = filename

    first_line = first_line_template % (display_filename, '[%s]' % message)
    return first_line


def console_progress_manager_factory(message, filename, size, steps=None,
                                     show_speed=False):
    if steps is None:
        steps = size

    first_line = _compute_optimal_first_line(message, filename)
    sys.stdout.write(first_line + "\n")

    left_align = 10
    # 2 for '[' and ']'
    width = len(first_line) - left_align - 2 - 1
    bar_template = "{0:>10}".format(human_bytes(size))
    bar_template += "%(label)s [%(bar)s] %(info)s"

    return ProgressBar(length=steps, bar_template=bar_template, width=width,
                       fill_char=".", show_speed=show_speed)


class _DummyProgressBar(object):
    def update(self, *a, **kw):
        pass

    def __enter__(self):
        return self

    def __exit__(self, *a, **kw):
        pass



#
# (C) Copyright 2014 Enthought, Inc., Austin, TX
# All right reserved.
#
# This file is open source software distributed according to the terms in
# LICENSE.txt
#

""" Utility functions to help with python wrapping.
"""
from __future__ import absolute_import

from ctypes import GetLastError, FormatError

from .pywintypes import error


def function_factory(
        function,  argument_types=None, return_type=None,
        error_checking=None):
    if argument_types is not None:
        function.argtypes = argument_types
    function.restype = return_type
    if error_checking is not None:
        function.errcheck = error_checking
    return function


def check_null_factory(func_name=None):
    def check_null(result, func, arguments, *args):
        if result is None:
            code = GetLastError()
            description = FormatError(code).strip()
            if func_name is None:
                raise error(code, func.__name__, description)
            else:
                raise error(code, func_name, description)
        return result
    return check_null

check_null = check_null_factory()


def check_zero_factory(func_name=None):
    def check_zero(result, func, arguments, *args):
        if result == 0:
            code = GetLastError()
            description = FormatError(code).strip()
            if func_name is None:
                raise error(code, func.__name__, description)
            else:
                raise error(code, func_name, description)
        return result
    return check_zero

check_zero = check_zero_factory()



"""
Copyright 2013 Dropbox, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import (
    absolute_import, print_function, division, unicode_literals
)

import re
import six

if six.PY2:
    try:
        from six import cStringIO as BufferIO
    except ImportError:
        from six import StringIO as BufferIO
else:
    from io import BytesIO as BufferIO

from collections import namedtuple, Sequence, Sized
from functools import wraps
from enstaller.vendor.requests.exceptions import ConnectionError
try:
    from enstaller.vendor.requests.packages.urllib3.response import HTTPResponse
except ImportError:
    from urllib3.response import HTTPResponse
if six.PY2:
    from urlparse import urlparse, parse_qsl
else:
    from urllib.parse import urlparse, parse_qsl


Call = namedtuple('Call', ['request', 'response'])


class CallList(Sequence, Sized):
    def __init__(self):
        self._calls = []

    def __iter__(self):
        return iter(self._calls)

    def __len__(self):
        return len(self._calls)

    def __getitem__(self, idx):
        return self._calls[idx]

    def add(self, request, response):
        self._calls.append(Call(request, response))

    def reset(self):
        self._calls = []


class RequestsMock(object):
    DELETE = 'DELETE'
    GET = 'GET'
    HEAD = 'HEAD'
    OPTIONS = 'OPTIONS'
    PATCH = 'PATCH'
    POST = 'POST'
    PUT = 'PUT'

    def __init__(self):
        self._calls = CallList()
        self.reset()

    def reset(self):
        self._urls = []
        self._calls.reset()

    def add(self, method, url, body='', match_querystring=False,
            status=200, adding_headers=None, stream=False,
            content_type='text/plain'):

        # ensure the url has a default path set if the url is a string
        if self._is_string(url) and url.count('/') == 2:
            url = url.replace('?', '/?', 1) if match_querystring \
                else url + '/'

        # body must be bytes
        if isinstance(body, six.text_type):
            body = body.encode('utf-8')

        self._urls.append({
            'url': url,
            'method': method,
            'body': body,
            'content_type': content_type,
            'match_querystring': match_querystring,
            'status': status,
            'adding_headers': adding_headers,
            'stream': stream,
        })

    def add_callback(self, method, url, callback, match_querystring=False,
                     content_type='text/plain'):

        self._urls.append({
            'url': url,
            'method': method,
            'callback': callback,
            'content_type': content_type,
            'match_querystring': match_querystring,
        })

    @property
    def calls(self):
        return self._calls

    def activate(self, func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            self.start()
            try:
                return func(*args, **kwargs)
            finally:
                self.stop()
                self.reset()
        return wrapped

    def _find_match(self, request):
        for match in self._urls:
            if request.method != match['method']:
                continue

            if not self._has_url_match(match, request.url):
                continue

            return match

        return None

    def _has_url_match(self, match, request_url):
        url = match['url']

        if self._is_string(url):
            if match['match_querystring']:
                return self._has_strict_url_match(url, request_url)
            else:
                url_without_qs = request_url.split('?', 1)[0]
                return url == url_without_qs
        elif isinstance(url, re._pattern_type) and url.match(request_url):
            return True
        else:
            return False

    def _has_strict_url_match(self, url, other):
        url_parsed = urlparse(url)
        other_parsed = urlparse(other)

        if url_parsed[:3] != other_parsed[:3]:
            return False

        url_qsl = sorted(parse_qsl(url_parsed.query))
        other_qsl = sorted(parse_qsl(other_parsed.query))
        return url_qsl == other_qsl

    def _is_string(self, s):
        return isinstance(s, (six.string_types, six.text_type))

    def _on_request(self, session, request, **kwargs):
        match = self._find_match(request)

        # TODO(dcramer): find the correct class for this
        if match is None:
            error_msg = 'Connection refused: {0}'.format(request.url)
            response = ConnectionError(error_msg)

            self._calls.add(request, response)
            raise response

        headers = {
            'Content-Type': match['content_type'],
        }

        if 'callback' in match:  # use callback
            status, r_headers, body = match['callback'](request)
            if isinstance(body, six.text_type):
                body = body.encode('utf-8')
            body = BufferIO(body)
            headers.update(r_headers)

        elif 'body' in match:
            if match['adding_headers']:
                headers.update(match['adding_headers'])
            status = match['status']
            body = BufferIO(match['body'])

        response = HTTPResponse(
            status=status,
            body=body,
            headers=headers,
            preload_content=False,
        )

        adapter = session.get_adapter(request.url)

        response = adapter.build_response(request, response)
        if not match.get('stream'):
            response.content  # NOQA

        self._calls.add(request, response)

        return response

    def start(self):
        import mock

        def unbound_on_send(session, requests, *a, **kwargs):
            return self._on_request(session, requests, *a, **kwargs)
        self._patcher = mock.patch('enstaller.vendor.requests.Session.send', unbound_on_send)
        self._patcher.start()

    def stop(self):
        self._patcher.stop()


# expose default mock namespace
_default_mock = RequestsMock()
__all__ = []
for __attr in (a for a in dir(_default_mock) if not a.startswith('_')):
    __all__.append(__attr)
    globals()[__attr] = getattr(_default_mock, __attr)



from enstaller.utils import PY_VER, canonical

from .dist_naming import split_eggname, filename_dist


class Req(object):
    """
    A requirement object is initalized by a requirement string. Attributes:
    name: the canonical project name
    version: the list of possible versions required
    strictness: the level of strictness
        0   nothing matters, anything matches
        1   only the name must match
        2   name and version must match
        3   name, version and build must match
    """
    def __init__(self, req_string):
        for c in '<>=,':
            assert c not in req_string, req_string
        lst = req_string.split()
        assert len(lst) <= 2, req_string
        self.strictness = 0
        self.name = self.version = self.build = None
        if lst:
            self.name = canonical(lst[0])
            self.strictness = 1
        if len(lst) == 2:
            tmp = lst[1]
            self.version = tmp.split('-')[0]
            self.strictness = 2 + bool('-' in tmp)
            if self.strictness ==  3:
                self.build = int(tmp.split('-')[1])

    def matches(self, spec):
        """
        Returns True if the spec of a distribution matches the requirement
        (self).  That is, the canonical name must match, and the version
        must be in the list of required versions.
        """
        #assert spec['metadata_version'] >= '1.1', spec
        if spec['python'] not in (None, PY_VER):
            return False
        if self.strictness == 0:
            return True
        if spec['cname'] != self.name:
            return False
        if self.strictness == 1:
            return True
        if spec['version'] != self.version:
            return False
        if self.strictness == 2:
            return True
        assert self.strictness == 3
        return spec['build'] == self.build

    def __str__(self):
        if self.strictness == 0:
            return ''
        res = self.name
        if self.version:
            res += ' %s' % self.version
        if self.build:
            res += '-%i' % self.build
        return res

    def __repr__(self):
        """
        return a canonical representation of the object
        """
        return 'Req(%r)' % str(self)

    def __eq__(self, other):
        return (self.name == other.name  and
                self.version == other.version  and
                self.build == other.build  and
                self.strictness == other.strictness)

    def __hash__(self):
        return (hash(self.strictness) ^ hash(self.name) ^
                hash(self.version) ^ hash(self.build))


def add_Reqs_to_spec(spec):
    """
    add the 'Reqs' key (which maps to the set of requirement objects),
    as well as the 'cname' key, to a spec dictionary
    """
    spec['Reqs'] = set(Req(s) for s in spec['packages'])
    spec['cname'] = canonical(spec['name'])


def spec_as_req(spec, strictness=3):
    """
    return a requirement object from a spec
    """
    assert 1 <= strictness <= 3
    req_string = spec['name']
    if strictness >= 2:
        req_string += ' %s' % spec['version']
    if strictness >= 3:
        req_string += '-%i' % spec['build']
    return Req(req_string)


def filename_as_req(filename, strictness=3):
    """
    return the filename of a distribution in terms of the a requirement object
    """
    name, version, build = split_eggname(filename)
    return spec_as_req(locals(), strictness)


def dist_as_req(dist, strictness=3):
    """
    Return the distribution in terms of the a requirement object.
    That is: What requirement gives me the distribution?
    """
    return filename_as_req(filename_dist(dist), strictness)



from __future__ import print_function, absolute_import

import os
import sys
import re
import bz2
import string
import zipfile
from collections import defaultdict
from os.path import basename, isfile, join, getmtime, getsize

from egginst.vendor.six import StringIO

from .dist_naming import is_valid_eggname
from .requirement import Req

from enstaller.utils import compute_md5


def parse_index(data):
    """
    Given the data of an index file, such as index-depend.txt, return a
    dictionary mapping the distribution names to the content of the
    corresponding section.
    """
    d = defaultdict(list)
    sep_pat = re.compile(r'==>\s*(\S+)\s*<==')
    for line in data.splitlines():
        m = sep_pat.match(line)
        if m:
            fn = m.group(1)
            continue
        d[fn].append(line.rstrip())

    res = {}
    for fn in d:
        res[fn] = '\n'.join(d[fn])
    return res


def data_from_spec(spec):
    """
    Given a spec dictionary, returns a the spec file as a well formed string.
    Also this function is a reference for meta-data version 1.1
    """
    str_None = str, type(None)
    for var, typ in [
        ('name', str), ('version', str), ('build', int),
        ('arch', str_None), ('platform', str_None), ('osdist', str_None),
        ('python', str_None), ('packages', list)]:
        assert isinstance(spec[var], typ), spec
        if isinstance(spec[var], str):
            s = spec[var]
            assert s == s.strip(), spec
            assert s != '', spec
    assert spec['build'] >= 0, spec

    cnames = set()
    for req_string in spec['packages']:
        r = Req(req_string)
        assert r.strictness >= 1
        cnames.add(r.name)
    # make sure no project is listed more than once
    assert len(cnames) == len(spec['packages'])

    lst = ["""\
metadata_version = '1.1'
name = %(name)r
version = %(version)r
build = %(build)i

arch = %(arch)r
platform = %(platform)r
osdist = %(osdist)r
python = %(python)r""" % spec]

    if spec['packages']:
        lst.append('packages = [')
        deps = spec['packages']
        for req in sorted(deps, key=string.lower):
            lst.append("  %r," % req)
        lst.append(']')
    else:
        lst.append('packages = []')

    lst.append('')
    return '\n'.join(lst)


def parse_data(data, index=False):
    """
    Given the content of a dependency spec file, return a dictionary mapping
    the variables to their values.

    If index is True, the MD5, size and mtime are also contained in the
    output dictionary.  It is an error these are missing in the input data.
    """
    spec = {}
    exec(data.replace('\r', ''), spec)
    assert spec['metadata_version'] >= '1.1', spec

    var_names = [ # these must be present
        'metadata_version', 'name', 'version', 'build',
        'arch', 'platform', 'osdist', 'python', 'packages']
    if index:
        # An index spec also has these
        var_names.extend(['md5', 'size'])
        assert isinstance(spec['md5'], str) and len(spec['md5']) == 32
        assert isinstance(spec['size'], int)
        if 'mtime' in spec:
            var_names.append('mtime')
        if 'commit' in spec:
            var_names.append('commit')

    res = {}
    for name in var_names:
        res[name] = spec[name]
    return res


def parse_depend_index(data):
    """
    Given the (uncompressed) data of index-depend.bz2, return a dict mapping
    each distname to a dict mapping variable names to their values.
    """
    d = parse_index(data)
    for fn in d:
        # convert the values from a text string (of the spec file) to a dict
        d[fn] = parse_data(d[fn], index=True)
    return d


def rawspec_from_dist(zip_path):
    """
    Returns the raw spec data, i.e. content of spec/depend as a string.
    """
    arcname = 'EGG-INFO/spec/depend'
    z = zipfile.ZipFile(zip_path)
    if arcname not in z.namelist():
        z.close()
        raise KeyError("arcname=%r not in zip-file %s" % (arcname, zip_path))
    data = z.read(arcname)
    z.close()
    return data


def spec_from_dist(zip_path):
    """
    Returns the spec dictionary from a zip-file distribution.
    """
    return parse_data(rawspec_from_dist(zip_path))


def commit_from_dist(zip_path):
    arcname = 'EGG-INFO/spec/__commit__'
    z = zipfile.ZipFile(zip_path)
    if arcname in z.namelist():
        res = 'commit = %r\n' % z.read(arcname).strip()
    else:
        res = ''
    z.close()
    return res


def index_section(zip_path):
    """
    Returns a section corresponding to the zip-file, which can be appended
    to an index.
    """
    return ('==> %s <==\n' % basename(zip_path) +
            'size = %i\n'  % getsize(zip_path) +
            'md5 = %r\n' % compute_md5(zip_path) +
            'mtime = %r\n' % getmtime(zip_path) +
            commit_from_dist(zip_path) +
            '\n' +
            rawspec_from_dist(zip_path) + '\n')


def write_txt_bz2(path, data):
    """
    Writes the 'data' to the file 'path' (which must end with '.txt'),
    as well as the bz2 compressed data to a file alongside 'data', where the
    txt extension is replaced by bz2.
    """
    fo = open(path, 'wt')
    fo.write(data)
    fo.close()
    assert path.endswith('.txt'), path
    fo = open(path[:-4] + '.bz2', 'wb')
    fo.write(bz2.compress(data))
    fo.close()


def update_index(dir_path, force=False, verbose=False):
    """
    Updates index-depend.txt and index-depend.bz2 in the directory specified.
    If index-depend.txt already exists, its content (which contains
    modification time stamps) is used to create the updated file.
    This can be disabled using the force option.
    """
    txt_path = join(dir_path, 'index-depend.txt')
    if verbose:
        print("Updating:", txt_path)

    if force or not isfile(txt_path):
        section = {}
    else:
        section = parse_index(open(txt_path).read())

    # since generating the new data may take a while, we first write to memory
    # and then write the file afterwards.
    faux = StringIO()
    for fn in sorted(os.listdir(dir_path), key=string.lower):
        if not fn.endswith('.egg'):
            continue
        if not is_valid_eggname(fn):
            print("WARNING: ignoring invalid egg name:", fn)
            continue
        path = join(dir_path, fn)
        if fn in section:
            spec = parse_data(section[fn], index=True)
            if spec.get('mtime') == getmtime(path):
                faux.write('==> %s <==\n' % fn)
                faux.write(section[fn] + '\n')
                continue
        faux.write(index_section(path))
        if verbose:
            sys.stdout.write('.')
            sys.stdout.flush()

    if verbose:
        print()
    write_txt_bz2(txt_path, faux.getvalue())
    faux.close()

    import enstaller.egg_meta
    enstaller.egg_meta.update_index(dir_path, force, verbose)



import logging

from ..util import properties
from ..util import XDG
from ..backend import KeyringBackend
from ..errors import (InitError, PasswordDeleteError,
    ExceptionRaisedContext)

try:
    import secretstorage
    import secretstorage.exceptions as exceptions
except ImportError:
    pass

log = logging.getLogger(__name__)

class Keyring(KeyringBackend):
    """Secret Service Keyring"""

    @properties.ClassProperty
    @classmethod
    @XDG.Preference('Gnome')
    def priority(cls):
        with ExceptionRaisedContext() as exc:
            secretstorage.__name__
        if exc:
            raise RuntimeError("SecretStorage required")
        if not hasattr(secretstorage, 'get_default_collection'):
            raise RuntimeError("SecretStorage 1.0 or newer required")
        try:
            bus = secretstorage.dbus_init()
            list(secretstorage.get_all_collections(bus))
        except exceptions.SecretServiceNotAvailableException as e:
            raise RuntimeError(
                "Unable to initialize SecretService: %s" % e)
        return 5

    def get_default_collection(self):
        bus = secretstorage.dbus_init()
        try:
            collection = secretstorage.get_default_collection(bus)
        except exceptions.SecretStorageException as e:
            raise InitError("Failed to create the collection: %s." % e)
        if collection.is_locked():
            collection.unlock()
            if collection.is_locked(): # User dismissed the prompt
                raise InitError("Failed to unlock the collection!")
        return collection

    def get_password(self, service, username):
        """Get password of the username for the service
        """
        collection = self.get_default_collection()
        items = collection.search_items(
            {"username": username, "service": service})
        for item in items:
            return item.get_secret().decode('utf-8')

    def set_password(self, service, username, password):
        """Set password for the username of the service
        """
        collection = self.get_default_collection()
        attributes = {
            "application": "python-keyring",
            "service": service,
            "username": username
            }
        label = "Password for '%s' on '%s'" % (username, service)
        collection.create_item(label, attributes, password, replace=True)

    def delete_password(self, service, username):
        """Delete the stored password (only the first one)
        """
        collection = self.get_default_collection()
        items = collection.search_items(
            {"username": username, "service": service})
        for item in items:
            return item.delete()
        raise PasswordDeleteError("No such password!")



import os
import struct
import sys


_DEFAULT_COLUMNS = 80
_DEFAULT_LINES = 25


def _posix_get_terminal_size():
    def ioctl_gwinsz(fd):
        try:
            import fcntl
            import termios
            cr = struct.unpack(
                'hh', fcntl.ioctl(fd, termios.TIOCGWINSZ, '1234'))
        except Exception:
            return
        return cr

    cr = ioctl_gwinsz(0) or ioctl_gwinsz(1) or ioctl_gwinsz(2)
    if not cr:
        try:
            fd = os.open(os.ctermid(), os.O_RDONLY)
            try:
                cr = ioctl_gwinsz(fd)
            finally:
                os.close(fd)
        except Exception:
            pass
    if not cr or not cr[0] or not cr[1]:
        cr = (os.environ.get('LINES', _DEFAULT_LINES),
              os.environ.get('COLUMNS', _DEFAULT_COLUMNS))
    return int(cr[1]), int(cr[0])


def get_terminal_size():
    """Returns the current size of the terminal as tuple in the form
    ``(width, height)`` in columns and rows.
    """
    # If shutil has get_terminal_size() (Python 3.3 and later) use that
    if sys.version_info >= (3, 3):
        import shutil
        shutil_get_terminal_size = getattr(shutil, 'get_terminal_size', None)
        if shutil_get_terminal_size:
            sz = shutil_get_terminal_size()
            return sz.columns, sz.lines

    # columns, lines are the working values
    try:
        columns = int(os.environ['COLUMNS'])
    except (KeyError, ValueError):
        columns = 0

    try:
        lines = int(os.environ['LINES'])
    except (KeyError, ValueError):
        lines = 0

    if columns <= 0 or lines <= 0:
        if os.name == "nt":
            from ._termui_win32 import get_terminal_size
            return get_terminal_size()
        else:
            return _posix_get_terminal_size()

    return columns, lines



from __future__ import print_function

import logging
import os
import sys

from os.path import isfile, join

from egginst.main import EggInst
from egginst.progress import dummy_progress_bar_factory

from enstaller.errors import EnpkgError, InvalidChecksum, NoSuchPackage
from enstaller.eggcollect import meta_dir_from_prefix
from enstaller.fetch import _DownloadManager
from enstaller.repository import (InstalledPackageMetadata, Repository,
                                  egg_name_to_name_version)

from enstaller.history import History
from enstaller.solver import Solver


_DEFAULT_MAX_RETRIES = 2

logger = logging.getLogger(__name__)


class _BaseAction(object):
    def __init__(self):
        self._is_canceled = False

    def __str__(self):
        return "{}: <{}>".format(self.__class__.__name__, self._egg)

    @property
    def is_canceled(self):
        return self._is_canceled

    def cancel(self):
        """ Cancel the action.

        Note: may not do anything for operations that cannot be safely
        canceled.
        """
        self._is_canceled = True

    def execute(self):
        """ Execute the given action."""

    def iter_execute(self):
        """ Iterator wich execute the action step by step when iterated
        over."""

    def __iter__(self):
        return self.iter_execute()


class FetchAction(_BaseAction):
    def __init__(self, egg, downloader, remote_repository, force=True,
                 progress_bar_factory=dummy_progress_bar_factory,
                 max_retries=_DEFAULT_MAX_RETRIES):
        super(FetchAction, self).__init__()
        self._downloader = downloader
        self._egg = egg
        self._force = force
        self._remote_repository = remote_repository

        self._progress_bar_factory = progress_bar_factory
        self._progress = None

        self._current_context = None
        self._retries = max_retries + 1

    def cancel(self):
        super(FetchAction, self).cancel()
        self._current_context.cancel()

    def progress_update(self, step):
        self._progress.update(step)

    def iter_execute(self):
        context = self._downloader.iter_fetch(self._egg, self._force)
        if not context.needs_to_download:
            return

        self._current_context = context

        name, version = egg_name_to_name_version(self._egg)
        package_metadata = self._remote_repository.find_package(name, version)
        progress = self._progress_bar_factory(package_metadata.key,
                                              package_metadata.size)

        with progress:
            self._progress = progress
            for chunk_size in context.iter_content():
                yield len(chunk_size)

    def execute(self):
        for i in range(self._retries):
            try:
                for chunk_size in self.iter_execute():
                    self.progress_update(chunk_size)
            except InvalidChecksum:
                if i >= self._retries - 1:
                    raise
            else:
                return


class InstallAction(_BaseAction):
    def __init__(self, egg, top_prefix, remote_repository,
                 top_installed_repository, installed_repository,
                 cache_directory,
                 progress_bar_factory=dummy_progress_bar_factory):
        super(InstallAction, self).__init__()

        self._egg = egg
        self._egg_path = os.path.join(cache_directory, self._egg)
        self._top_prefix = top_prefix
        self._remote_repository = remote_repository
        self._top_installed_repository = top_installed_repository
        self._installed_repository = installed_repository

        self._progress_factory = progress_bar_factory
        self._progress = None

    def progress_update(self, step):
        self._progress.update(step)

    def _extract_extra_info(self):
        name, version = egg_name_to_name_version(self._egg)
        package = self._remote_repository.find_package(name, version)
        return package.s3index_data

    def iter_execute(self):
        extra_info = self._extract_extra_info()

        installer = EggInst(self._egg_path, prefix=self._top_prefix)

        self._progress = self._progress_factory(installer.fn,
                                                installer.installed_size)

        with self._progress:
            for step in installer.install_iterator(extra_info):
                yield step

        self._post_install()

    def execute(self):
        for currently_extracted_size in self.iter_execute():
            self.progress_update(currently_extracted_size)

    def _post_install(self):
        name, _ = egg_name_to_name_version(self._egg_path)
        meta_dir = meta_dir_from_prefix(self._top_prefix, name)
        package = InstalledPackageMetadata.from_meta_dir(meta_dir)

        self._top_installed_repository.add_package(package)
        self._installed_repository.add_package(package)


class RemoveAction(_BaseAction):
    def __init__(self, egg, top_prefix, top_installed_repository,
                 installed_repository,
                 progress_bar_factory=dummy_progress_bar_factory):
        super(RemoveAction, self).__init__()
        self._egg = egg
        self._top_prefix = top_prefix

        self._top_installed_repository = top_installed_repository
        self._installed_repository = installed_repository

        self._progress_factory = progress_bar_factory
        self._progress = None

    def progress_update(self, step):
        self._progress.update(step)

    def iter_execute(self):
        installer = EggInst(self._egg, self._top_prefix, False)
        remover = installer._egginst_remover
        if not remover.is_installed:
            logger.error("Error: can't find meta data for: %r", remover.cname)
            return
        name, version = egg_name_to_name_version(self._egg)
        package = self._top_installed_repository.find_package(name, version)

        self._progress = self._progress_factory(installer.fn,
                                                remover.installed_size,
                                                len(remover.files))

        with self._progress:
            for filename in remover.remove_iterator():
                yield 1

        self._top_installed_repository.delete_package(package)
        self._installed_repository.delete_package(package)

    def execute(self):
        for n in self.iter_execute():
            self.progress_update(n)


class ProgressBarContext(object):
    def __init__(self, default_progress_bar_factory, **kw):
        """
        A simple object holding progress bar factories to be used by actions.

        Parameters
        ----------
        default_progress_bar_factory : callable
            The default progress bar factory
        **kw :
            action_name -> callable pairs. For each specified action, this
            specific callable will be used instead of the default factory.

        .. note:: each callable signature must be (label, filename, size)
        """
        self.default_progress_bar_factory = default_progress_bar_factory

        self.fetch_progress_factory = kw.get("fetch",
                                             default_progress_bar_factory)
        self.install_progress_factory = kw.get("install",
                                               default_progress_bar_factory)
        self.remove_progress_factory = kw.get("remove",
                                              default_progress_bar_factory)

    def fetch_progress(self, filename, size):
        return self.fetch_progress_factory("fetching", filename, size)

    def install_progress(self, filename, size):
        return self.install_progress_factory("installing egg", filename, size)

    def remove_progress(self, filename, size, steps):
        return self.remove_progress_factory("removing egg", filename, size,
                                            steps)


class _ExecuteContext(object):
    def __init__(self, actions, enpkg, progress_bar_context, force=False,
                 max_retries=_DEFAULT_MAX_RETRIES):
        self._top_prefix = enpkg.top_prefix
        self._actions = actions
        self._remote_repository = enpkg._remote_repository
        self._enpkg = enpkg
        self._force = force

        self._pbar_context = progress_bar_context

        self._max_retries = max_retries

    def _action_factory(self, action):
        opcode, egg = action

        if opcode.startswith('fetch'):
            return FetchAction(egg, self._enpkg._downloader,
                               self._remote_repository, self._force,
                               self._pbar_context.fetch_progress,
                               self._max_retries)
        elif opcode.startswith("install"):
            return InstallAction(egg, self._enpkg.top_prefix,
                                 self._enpkg._remote_repository,
                                 self._enpkg._top_installed_repository,
                                 self._enpkg._installed_repository,
                                 self._enpkg._downloader.cache_directory,
                                 self._pbar_context.install_progress)
        elif opcode.startswith("remove"):
            return RemoveAction(egg, self._enpkg.top_prefix,
                                self._enpkg._top_installed_repository,
                                self._enpkg._installed_repository,
                                self._pbar_context.remove_progress)
        else:
            raise ValueError("Unknown opcode: {0!r}".format(opcode))

    def __iter__(self):
        with History(self._top_prefix):
            for action in self._actions:
                logger.info('\t' + str(action))
                yield self._action_factory(action)


class Enpkg(object):
    """ This is main interface for using enpkg, it is used by the CLI.
    Arguments for object creation:

    Parameters
    ----------
    repository : Repository
        This is the remote repository which enpkg will use to resolve
        dependencies.
    session : Session
        The session used to connect to the remote server (fetching indices,
        eggs, etc...).
    prefixes : list of paths -- default: [sys.prefix]
        Each path, is an install "prefix" (such as, e.g. /usr/local) in which
        things get installed. Eggs are installed or removed from the first
        prefix in the list.
    progress_context : ProgressBarContext
        If specified, will be used for progress bar management across all
        executed actions. If None, use dummy (do nothing) progress bars.
    max_retries : int
        Maximum number of retries to fetch an egg when checksum mismatchs
        occur.
    """
    def __init__(self, remote_repository, session,
                 prefixes=[sys.prefix], progress_context=None,
                 force=False, max_retries=_DEFAULT_MAX_RETRIES):
        self.prefixes = prefixes
        self.top_prefix = prefixes[0]

        self._remote_repository = remote_repository

        self._installed_repository = Repository._from_prefixes(self.prefixes)
        self._top_installed_repository = \
            Repository._from_prefixes([self.top_prefix])

        self._session = session
        self._downloader = _DownloadManager(session, remote_repository)

        self._progress_context = progress_context or \
            ProgressBarContext(dummy_progress_bar_factory)

        self._force = force
        self.max_retries = max_retries

    def _solver_factory(self, mode='recur', force=False, forceall=False):
        solver = Solver(self._remote_repository,
                        self._top_installed_repository,
                        mode, force, forceall)
        return solver

    def execute_context(self, actions):
        return _ExecuteContext(actions, self, self._progress_context,
                               self._force)

    def execute(self, actions):
        """
        Execute the given set of actions.

        This method is only meant to be called with actions created by the
        *_actions methods below.

        Parameters
        ----------
        actions : list
            List of (opcode, egg) pairs, as returned by the *_actions from
            Solver.
        """
        logger.info("Enpkg.execute: %d", len(actions))
        for action in self.execute_context(actions):
            action.execute()

    def revert_actions(self, arg):
        """
        Calculate the actions necessary to revert to a given state, the
        argument may be one of:
          * complete set of eggs, i.e. a set of egg file names
          * revision number (negative numbers allowed)
        """
        h = History(self.top_prefix)
        h.update()
        if isinstance(arg, set):
            state = arg
        else:
            try:
                rev = int(arg)
            except (TypeError, ValueError):
                raise EnpkgError("Invalid argument: integer expected, "
                                 "got: {0!r}".format(arg))
            try:
                state = h.get_state(rev)
            except IndexError:
                raise EnpkgError("Error: no such revision: %r" % arg)

        curr = h.get_state()
        if state == curr:
            return []

        res = []
        for egg in curr - state:
            if egg.startswith('enstaller'):
                continue
            res.append(('remove', egg))

        for egg in state - curr:
            if egg.startswith('enstaller'):
                continue
            if not isfile(join(self._downloader.cache_directory, egg)):
                eggname, version = egg_name_to_name_version(egg)
                try:
                    self._remote_repository.find_package(eggname, version)
                    res.append(('fetch_0', egg))
                except NoSuchPackage:
                    raise EnpkgError("cannot revert -- missing %r" % egg)
            res.append(('install', egg))
        return res

    def get_history(self):
        """
        return a history (h) object with this Enpkg instance prefix.
        """
        # FIXME: only used by canopy
        return History(self.top_prefix)



import platform
import subprocess
import re
import binascii

from ..backend import KeyringBackend
from ..errors import PasswordSetError
from ..errors import PasswordDeleteError
from ..util import properties
from ..py27compat import unicode_str


class SecurityCommand(unicode_str):
    """
    A string suitable for passing as the 'command' parameter to the
    OS X 'security' command.
    """
    def __new__(cls, cmd, store='generic'):
        cmd = '%(cmd)s-%(store)s-password' % vars()
        return super(SecurityCommand, cls).__new__(cls, cmd)


class Keyring(KeyringBackend):
    """Mac OS X Keychain"""

    # regex for extracting password from security call
    password_regex = re.compile("""password:\s*(?:0x(?P<hex>[0-9A-F]+)\s*)?"""
                                """(?:"(?P<pw>.*)")?""")
    store = 'generic'

    @properties.ClassProperty
    @classmethod
    def priority(cls):
        """
        Preferred for all OS X environments.
        """
        if platform.system() != 'Darwin':
            raise RuntimeError("OS X required")
        return 5

    def set_password(self, service, username, password):
        if username is None:
            username = ''
        set_error = PasswordSetError("Can't store password in keychain")
        try:
            # set up the call for security.
            cmd = [
                'security',
                SecurityCommand('add', self.store),
                '-a', username,
                '-s', service,
                '-w', password,
                '-U',
            ]
            call = subprocess.Popen(cmd, stderr=subprocess.PIPE,
                stdout=subprocess.PIPE)
            stdoutdata, stderrdata = call.communicate()
            code = call.returncode
            # check return code.
            if code is not 0:
                raise set_error
        except:
            raise set_error

    def get_password(self, service, username):
        if username is None:
            username = ''
        try:
            # set up the call to security.
            cmd = [
                'security',
                SecurityCommand('find', self.store),
                '-g',
                '-a', username,
                '-s', service,
            ]
            call = subprocess.Popen(cmd, stderr=subprocess.PIPE,
                stdout=subprocess.PIPE)
            stdoutdata, stderrdata = call.communicate()
            code = call.returncode
            if code is not 0:
                raise OSError("Can't fetch password from system")
            output = stderrdata.decode()
            # check for empty password.
            if output == 'password: \n':
                return ''
            # search for special password pattern.
            matches = Keyring.password_regex.search(output)
            if matches:
                group_dict = matches.groupdict()
                hex = group_dict.get('hex')
                pw = group_dict.get('pw')
                if hex:
                    # it's a weird hex password, decode it.
                    return unicode_str(binascii.unhexlify(hex), 'utf-8')
                else:
                    # it's a normal password, send it back.
                    return pw
            # nothing was found, it doesn't exist.
        except:
            pass

    def delete_password(self, service, username):
        del_error = PasswordDeleteError("Can't delete password in keychain")
        if username is None:
            username = ''
        try:
            cmd = [
                'security',
                SecurityCommand('delete', self.store),
                '-a', username,
                '-s', service,
            ]
            # set up the call for security.
            call = subprocess.Popen(cmd, stderr=subprocess.PIPE,
                stdout=subprocess.PIPE)
            stdoutdata, stderrdata = call.communicate()
            code = call.returncode
            # check return code.
            if code is not 0:
                raise del_error
        except:
            raise del_error



import sys
import base64
import platform

from ..py27compat import unicode_str
from ..util import escape, properties
from ..backend import KeyringBackend
from ..errors import PasswordDeleteError, ExceptionRaisedContext
from . import file

try:
    from enstaller.vendor.win32ctypes import pywintypes
    from enstaller.vendor.win32ctypes import win32cred
except (ImportError, ValueError, KeyError):
    # ValueError may be raised on non windows machines because of ctypes
    # KeyError may be raised on non windows machines because of ctypes with pypy
    pass

try:
    import winreg
except ImportError:
    try:
        # Python 2 compatibility
        import _winreg as winreg
    except ImportError:
        pass

try:
    from . import _win_crypto
except ImportError:
    pass

def has_pywin32():
    """
    Does this environment have pywin32?
    Should return False even when Mercurial's Demand Import allowed import of
    win32cred.
    """
    with ExceptionRaisedContext() as exc:
        win32cred.__name__
    return not bool(exc)

def has_wincrypto():
    """
    Does this environment have wincrypto?
    Should return False even when Mercurial's Demand Import allowed import of
    _win_crypto, so accesses an attribute of the module.
    """
    with ExceptionRaisedContext() as exc:
        _win_crypto.__name__
    return not bool(exc)

class EncryptedKeyring(file.BaseKeyring):
    """
    A File-based keyring secured by Windows Crypto API.
    """

    @properties.ClassProperty
    @classmethod
    def priority(self):
        """
        Preferred over file.EncryptedKeyring but not other, more sophisticated
        Windows backends.
        """
        if not platform.system() == 'Windows':
            raise RuntimeError("Requires Windows")
        return .8

    filename = 'wincrypto_pass.cfg'

    def encrypt(self, password):
        """Encrypt the password using the CryptAPI.
        """
        return _win_crypto.encrypt(password)

    def decrypt(self, password_encrypted):
        """Decrypt the password using the CryptAPI.
        """
        return _win_crypto.decrypt(password_encrypted)


class WinVaultKeyring(KeyringBackend):
    """
    WinVaultKeyring stores encrypted passwords using the Windows Credential
    Manager.

    Requires pywin32

    This backend does some gymnastics to simulate multi-user support,
    which WinVault doesn't support natively. See
    https://bitbucket.org/kang/python-keyring-lib/issue/47/winvaultkeyring-only-ever-returns-last#comment-731977
    for details on the implementation, but here's the gist:

    Passwords are stored under the service name unless there is a collision
    (another password with the same service name but different user name),
    in which case the previous password is moved into a compound name:
    {username}@{service}
    """

    @properties.ClassProperty
    @classmethod
    def priority(cls):
        """
        If available, the preferred backend on Windows.
        """
        if not has_pywin32():
            raise RuntimeError("Requires Windows and pywin32")
        return 5

    @staticmethod
    def _compound_name(username, service):
        return escape.u('%(username)s@%(service)s') % vars()

    def get_password(self, service, username):
        # first attempt to get the password under the service name
        res = self._get_password(service)
        if not res or res['UserName'] != username:
            # It wasn't found so attempt to get it with the compound name
            res = self._get_password(self._compound_name(username, service))
        if not res:
            return None
        blob = res['CredentialBlob']
        return blob.decode('utf-16')

    def _get_password(self, target):
        try:
            res = win32cred.CredRead(
                Type=win32cred.CRED_TYPE_GENERIC,
                TargetName=target,
            )
        except pywintypes.error:
            e = sys.exc_info()[1]
            if e.winerror == 1168 and e.funcname == 'CredRead': # not found
                return None
            raise
        return res

    def set_password(self, service, username, password):
        existing_pw = self._get_password(service)
        if existing_pw:
            # resave the existing password using a compound target
            existing_username = existing_pw['UserName']
            target = self._compound_name(existing_username, service)
            self._set_password(target, existing_username,
                existing_pw['CredentialBlob'].decode('utf-16'))
        self._set_password(service, username, unicode_str(password))

    def _set_password(self, target, username, password):
        credential = dict(Type=win32cred.CRED_TYPE_GENERIC,
                          TargetName=target,
                          UserName=username,
                          CredentialBlob=password,
                          Comment="Stored using python-keyring",
                          Persist=win32cred.CRED_PERSIST_ENTERPRISE)
        win32cred.CredWrite(credential, 0)

    def delete_password(self, service, username):
        compound = self._compound_name(username, service)
        deleted = False
        for target in service, compound:
            existing_pw = self._get_password(target)
            if existing_pw and existing_pw['UserName'] == username:
                deleted = True
                self._delete_password(target)
        if not deleted:
            raise PasswordDeleteError(service)

    def _delete_password(self, target):
        win32cred.CredDelete(
            Type=win32cred.CRED_TYPE_GENERIC,
            TargetName=target,
        )


class RegistryKeyring(KeyringBackend):
    """
    RegistryKeyring is a keyring which use Windows CryptAPI to encrypt
    the user's passwords and store them under registry keys
    """

    @properties.ClassProperty
    @classmethod
    def priority(self):
        """
        Preferred on Windows when pywin32 isn't installed
        """
        if platform.system() != 'Windows':
            raise RuntimeError("Requires Windows")
        if not has_wincrypto():
            raise RuntimeError("Requires ctypes")
        return 2

    def get_password(self, service, username):
        """Get password of the username for the service
        """
        try:
            # fetch the password
            key = r'Software\%s\Keyring' % service
            hkey = winreg.OpenKey(winreg.HKEY_CURRENT_USER, key)
            password_saved = winreg.QueryValueEx(hkey, username)[0]
            password_base64 = password_saved.encode('ascii')
            # decode with base64
            password_encrypted = base64.decodestring(password_base64)
            # decrypted the password
            password = _win_crypto.decrypt(password_encrypted).decode('utf-8')
        except EnvironmentError:
            password = None
        return password

    def set_password(self, service, username, password):
        """Write the password to the registry
        """
        # encrypt the password
        password_encrypted = _win_crypto.encrypt(password.encode('utf-8'))
        # encode with base64
        password_base64 = base64.encodestring(password_encrypted)
        # encode again to unicode
        password_saved = password_base64.decode('ascii')

        # store the password
        key_name = r'Software\%s\Keyring' % service
        hkey = winreg.CreateKey(winreg.HKEY_CURRENT_USER, key_name)
        winreg.SetValueEx(hkey, username, 0, winreg.REG_SZ, password_saved)

    def delete_password(self, service, username):
        """Delete the password for the username of the service.
        """
        try:
            key_name = r'Software\%s\Keyring' % service
            hkey = winreg.OpenKey(winreg.HKEY_CURRENT_USER, key_name, 0,
                winreg.KEY_ALL_ACCESS)
            winreg.DeleteValue(hkey, username)
        except WindowsError:
            e = sys.exc_info()[1]
            raise PasswordDeleteError(e)



import argparse
import os
import os.path
import re
import shutil
import sys
import tempfile

import okonomiyaki

try:
    version_info = okonomiyaki.__version_info__
except AttributeError:
    raise RuntimeError("You need okonomiyaki >= 0.9.0 to use repack.")
else:
    if version_info < (0, 9):
        raise RuntimeError("You need okonomiyaki >= 0.9.0 to use repack.")

from okonomiyaki.errors import OkonomiyakiError
from okonomiyaki.file_formats import (
    Dependencies, EggBuilder, EggMetadata, PackageInfo, Requirement,
    is_egg_name_valid,
)
from okonomiyaki.file_formats.setuptools_egg import (
    _UNSPECIFIED, SetuptoolsEggMetadata, parse_filename
)
from okonomiyaki.platforms import EPDPlatform

from egginst.eggmeta import SPEC_DEPEND_KEYS
from egginst.utils import samefile
from egginst._zipfile import ZipFile
from enstaller.errors import EnstallerException
from enstaller.versions.enpkg import EnpkgVersion
from enstaller.versions.pep386 import suggest_normalized_version


ENDIST_DAT = "endist.dat"
# Whitelist of keys considered when exec`ing the endist.dat to update
# spec/depend
ACCEPTED_ENDIST_SPEC_KEYS = SPEC_DEPEND_KEYS
# Whitelist of keys considered when exec`ing the endist.dat to update
# the egg content
ACCEPTED_ENDIST_EGG_KEYS = ("add_files", )
# List of endist keys not yet supported by this tool
UNSUPPORTED_ENDIST_KEYS = ("app_icon_file", "svn_rev", "svn_rev_init",
                           "no_pyc")

_SETUPTOOLS_TYPE = "setuptools"
_EGGINST_TYPE = "egginst"


class InvalidVersion(EnstallerException):
    def __init__(self, *args, **kw):
        self.version, = args
        super(InvalidVersion, self).__init__(*args, **kw)

    def __str__(self):
        suggested = suggest_normalized_version(self.version)
        template = ("The given version '{0}' does not follow PEP 386. Please "
                    "change the egg version to a valid format")
        if suggested is not None:
            template += " (e.g. '{0}').".format(suggested)
        else:
            template += "."
        return template.format(self.version)


def _looks_like_enthought_egg(path):
    """ Returns True if the give file looks like an enthought egg.
    """
    filename = os.path.basename(path)

    if is_egg_name_valid(filename):
        with ZipFile(path) as zp:
            try:
                zp.getinfo("EGG-INFO/spec/depend")
                return True
            except KeyError:
                pass

    return False


def _looks_like_setuptools_egg(path):
    filename = os.path.basename(path)

    try:
        parse_filename(filename)
        return True
    except OkonomiyakiError:
        return False


def _get_spec_data(source_egg_path, build_number, platform_string=None,
                   python=_UNSPECIFIED, abi_tag=_UNSPECIFIED):
    if _looks_like_setuptools_egg(source_egg_path):
        parsed_platform_string = parse_filename(source_egg_path)[3]
        if parsed_platform_string is not None and platform_string is None:
            msg = ("Platform-specific egg detected (platform string is "
                   "{0!r}), you *must* specify the platform.")
            raise EnstallerException(msg.format(parsed_platform_string))

        if platform_string is not None:
            platform = EPDPlatform.from_epd_string(platform_string)
        else:
            platform = None
        metadata = SetuptoolsEggMetadata.from_egg(
            source_egg_path, platform, python, abi_tag
        )
        egg_basename = metadata.name
        version = str(metadata.version)
        dependencies = []
    elif _looks_like_enthought_egg(source_egg_path):
        metadata = EggMetadata.from_egg(source_egg_path)
        # The name as used in spec/depend and endist.dat is the so-called
        # egg basename (i.e. not normalied to lower case)
        egg_basename = metadata.egg_basename
        version = metadata.upstream_version
        dependencies = metadata.runtime_dependencies
    else:
        msg = "Unrecognized format: {0!r}".format(source_egg_path)
        raise EnstallerException(msg)

    data = {"build": build_number, "packages": dependencies,
            "name": egg_basename, "version": version}

    if os.path.exists(ENDIST_DAT):
        data.update(_parse_endist_for_spec_depend(ENDIST_DAT))

    return data, metadata


def _raw_packages_to_requirements(packages):
    return tuple(Requirement.from_spec_string(s) for s in packages)


def _get_spec(source_egg_path, build_number, platform_string=None,
              python=_UNSPECIFIED, abi_tag=_UNSPECIFIED):
    data, metadata = _get_spec_data(source_egg_path, build_number,
                                    platform_string, python, abi_tag)

    if platform_string is None:
        try:
            epd_platform = EPDPlatform.from_running_system()
        except OkonomiyakiError as e:
            msg = "Could not guess platform from system (original " \
                  "error was {0!r}). " \
                  "You may have to specify the platform explicitly."
            raise EnstallerException(msg.format(e))
    else:
        epd_platform = EPDPlatform.from_epd_string(platform_string)

    raw_name = data["name"]
    version = EnpkgVersion.from_upstream_and_build(data["version"],
                                                   data["build"])
    if version.upstream.is_worked_around:
        raise InvalidVersion(str(version.upstream))

    dependencies = Dependencies(runtime=data["packages"])
    pkg_info = PackageInfo.from_egg(source_egg_path)
    return EggMetadata(raw_name, version, epd_platform, metadata.python_tag,
                       metadata.abi_tag, dependencies, pkg_info,
                       metadata.summary)


def _parse_endist_for_egg_content(path):
    data = _parse_endist(path)
    return dict((k, data[k]) for k in data if k in ACCEPTED_ENDIST_EGG_KEYS)


def _parse_endist_for_spec_depend(path):
    data = _parse_endist(path)
    data["packages"] = _raw_packages_to_requirements(data.get("packages", []))
    return dict((k, data[k]) for k in data if k in ACCEPTED_ENDIST_SPEC_KEYS)


def _parse_endist(path):
    with open(path) as fp:
        globals_dict = {}
        local_dict = {}
        exec(fp.read(), local_dict, globals_dict)

    for k in globals_dict:
        if k in UNSUPPORTED_ENDIST_KEYS:
            msg = "key {0!r} not yet supported in {1!r}".format(k, ENDIST_DAT)
            raise NotImplementedError(msg)
    return globals_dict


def _add_files(z, dir_path, regex_string, archive_dir):
    r = re.compile(regex_string)
    archive_dir = archive_dir.strip('/')

    print("dir_path: {0!r}".format(dir_path))
    print("rx: {0!r}".format(regex_string))
    print("archive_dir: {0!r}".format(archive_dir))

    arcnames = set()

    for filename in os.listdir(dir_path):
        path = os.path.join(dir_path, filename)
        if not (r.match(filename) and os.path.isfile(path)):
            continue

        print("\tfn={0!r}".format(filename))

        arcname = archive_dir + '/' + filename
        if os.path.islink(path):
            msg = "Soft link support not yet implemented\n"
            raise NotImplementedError(msg)
        elif os.path.isfile(path):
            z.add_file_as(path, arcname)
            arcnames.add(arcname)
        else:
            raise EnstallerException("Neiher link nor file:" % path)

    return arcnames


def repack(source_egg_path, build_number=1, platform_string=None,
           python=_UNSPECIFIED, abi_tag=_UNSPECIFIED):
    metadata = _get_spec(
        source_egg_path, build_number, platform_string, python, abi_tag
    )

    parent_dir = os.path.dirname(os.path.abspath(source_egg_path))
    target_egg_path = os.path.join(parent_dir, metadata.egg_name)

    if os.path.exists(target_egg_path) and \
            samefile(source_egg_path, target_egg_path):
        msg = "source and repack-ed egg are the same file: {0!r}. Inplace " \
              "mode not yet implemented."
        raise EnstallerException(msg.format(source_egg_path))
    # XXX: implement endist.dat/app handling

    print(20 * '-' + '\n' + metadata.spec_depend_string + 20 * '-')

    with ZipFile(source_egg_path) as source:
        tempdir = tempfile.mkdtemp()
        try:
            with EggBuilder(metadata, cwd=parent_dir) as target:
                if os.path.exists(ENDIST_DAT):
                    data = _parse_endist_for_egg_content(ENDIST_DAT)
                    for entry in data.get("add_files", []):
                        _add_files(target, entry[0], entry[1], entry[2])
                files_to_skip = set(target._fp.namelist())

                for f in source.namelist():
                    if f not in files_to_skip:
                        source_path = source.extract(f, tempdir)
                        target.add_file_as(source_path, f)
        finally:
            shutil.rmtree(tempdir)


def main(argv=None):
    argv = argv or sys.argv[1:]

    p = argparse.ArgumentParser("Egg repacker.")
    p.add_argument("egg", help="Path to the egg to repack.")
    p.add_argument("-b", "--build", dest="build_number",
                   help="Build number (default is %(default)s)",
                   default=1, type=int)
    p.add_argument("-a", dest="platform_string",
                   help="Legacy epd platform string (e.g. 'rh5-32'). "
                        "Will be guessed if not specified.")
    p.add_argument("--abi", dest="abi_tag",
                   help="PEP425 abi tag. Will be guessed if not specified."
                        "Note: the tag format will not be validated.")
    ns = p.parse_args(argv)

    abi_tag = ns.abi_tag or _UNSPECIFIED
    repack(ns.egg, ns.build_number, ns.platform_string, abi_tag=abi_tag)


if __name__ == "__main__":  # pragma: no cover
    main()



from __future__ import print_function

import logging
import os
from os.path import dirname, isdir, join

from egginst.utils import rm_rf

logger = logging.getLogger(__name__)


# XXX: kept for legacy reason, DO NOT REMOVE.
# This code is used during enstaller inplace update.
def create(egg_inst):
    if egg_inst.cname != "enstaller":
        msg = "Legacy dummy code used for an egg not called enstaller !"
        raise RuntimeError(msg)


def create_link(arcname, link, prefix):
    usr = 'EGG-INFO/usr/'
    assert arcname.startswith(usr), arcname
    dst = join(prefix, arcname[len(usr):])

    # Create the destination directory if it does not exist.  In most cases
    # it will exist, but you never know.
    if not isdir(dirname(dst)):
        os.makedirs(dirname(dst))

    rm_rf(dst)
    logger.info("Creating: %s (link to %s)", dst, link)
    os.symlink(link, dst)
    return dst



from __future__ import absolute_import

import sys
import textwrap

if sys.version_info < (2, 7):
    import unittest2 as unittest
    # FIXME: this looks quite fishy. On 2.6, with unittest2, the assertRaises
    # context manager does not contain the actual exception object ?
    def exception_code(ctx):
        return ctx.exception
else:
    import unittest
    def exception_code(ctx):
        return ctx.exception.code

import mock

from egginst.testing_utils import ControlledEnv
from enstaller.errors import EXIT_ABORTED
from enstaller.main import main_noexc
from enstaller.tests.common import mock_print

from .common import mock_install_req, set_env_vars, without_any_configuration

class TestEnstallerMainActions(unittest.TestCase):
    @without_any_configuration
    def test_print_version(self):
        # XXX: this is lousy test: we'd like to at least ensure we're printing
        # the correct version, but capturing the stdout is a bit tricky. Once
        # we replace print by proper logging, we should be able to do better.
        with self.assertRaises(SystemExit) as e:
            main_noexc(["--version"])
        self.assertEqual(e.exception.code, 0)

    @without_any_configuration
    def test_help_runs_and_exits_correctly(self):
        with self.assertRaises(SystemExit) as e:
            main_noexc(["--help"])
        self.assertEqual(e.exception.code, 0)

    @without_any_configuration
    def test_print_env(self):
        with self.assertRaises(SystemExit) as e:
            main_noexc(["--env"])
        self.assertEqual(e.exception.code, 0)

    @without_any_configuration
    @mock_install_req
    def test_verbose(self):
        with self.assertRaises(SystemExit) as e:
            main_noexc(["--verbose", "--env"])
        self.assertEqual(e.exception.code, 0)

    @without_any_configuration
    def test_ctrl_c_handling(self):
        with mock.patch("enstaller.main.main", side_effect=KeyboardInterrupt):
            with self.assertRaises(SystemExit) as e:
                main_noexc()
                self.assertEqual(exception_code(e), EXIT_ABORTED)

    @unittest.expectedFailure
    @without_any_configuration
    def test_crash_handling_default(self):
        r_output = textwrap.dedent("""\
        enstaller: Error: enstaller crashed (uncaught exception {0}: Exception()).
        Please report this on enstaller issue tracker:
            http://github.com/enthought/enstaller/issues
        You can get a full traceback by setting the ENSTALLER_DEBUG environment variable
        """.format(Exception))
        env = ControlledEnv(["ENSTALLER_DEBUG"])
        with mock.patch("enstaller.main.os.environ", env):
            with mock.patch("enstaller.main.main", side_effect=Exception):
                with mock_print() as mocked_print:
                    with self.assertRaises(SystemExit) as e:
                        main_noexc()
                    self.assertEqual(exception_code(e), 1)
                self.assertMultiLineEqual(mocked_print.value, r_output)

    @without_any_configuration
    def test_crash_handling_debug(self):
        with set_env_vars(ENSTALLER_DEBUG="1"):
            with mock.patch("enstaller.main.main", side_effect=Exception):
                with mock_print() as mocked_print:
                    with self.assertRaises(Exception):
                        main_noexc()
                self.assertMultiLineEqual(mocked_print.value, "")



import os
import os.path
import shutil
import sys
import tempfile

if sys.version_info < (2, 7):
    import unittest2 as unittest
else:
    import unittest

import mock
import testfixtures

from egginst.main import (
    EggInst, get_installed, is_in_legacy_egg_info, main,
    should_copy_in_egg_info)
from egginst.testing_utils import assert_same_fs
from egginst.utils import makedirs, zip_write_symlink
from egginst._zipfile import ZipFile

from egginst import eggmeta

from .common import (DUMMY_EGG, DUMMY_EGG_WITH_APPINST,
                     DUMMY_EGG_WITH_ENTRY_POINTS, DUMMY_EGG_METADATA_FILES,
                     LEGACY_EGG_INFO_EGG, LEGACY_EGG_INFO_EGG_METADATA_FILES,
                     NOSE_1_3_0, PYTHON_VERSION, STANDARD_EGG,
                     STANDARD_EGG_METADATA_FILES, SUPPORT_SYMLINK,
                     VTK_EGG_DEFERRED_SOFTLINK, mkdtemp)


def _create_egg_with_symlink(filename, name):
    with ZipFile(filename, "w") as fp:
        fp.writestr("EGG-INFO/usr/include/foo.h", "/* header */")
        zip_write_symlink(fp, "EGG-INFO/usr/HEADERS", "include")


def _create_dummy_enstaller_egg(prefix):
    path = os.path.join(prefix, "enstaller-4.8.0-1.egg")
    with ZipFile(path, "w") as fp:
        # We write a dummy file as empty zip files are not properly handled by
        # zipfile in python 2.6
        fp.writestr("EGG-INFO/dummy", "")
    return path


class TestEggInst(unittest.TestCase):
    def setUp(self):
        self.base_dir = tempfile.mkdtemp()
        makedirs(self.base_dir)
        self.prefix = os.path.join(self.base_dir, "prefix")

    def tearDown(self):
        shutil.rmtree(self.base_dir)

    @unittest.skipIf(not SUPPORT_SYMLINK,
                     "this platform does not support symlink")
    def test_symlink(self):
        """Test installing an egg with softlink in it."""
        # Given
        if sys.platform == "win32":
            incdir = os.path.join(self.prefix, "EGG-INFO", "foo", "usr",
                                  "include")
            header = os.path.join(incdir, "foo.h")
            link = os.path.join(self.prefix, "EGG-INFO", "foo", "usr",
                                "HEADERS")
        else:
            incdir = os.path.join(self.prefix, "include")
            header = os.path.join(incdir, "foo.h")
            link = os.path.join(self.prefix, "HEADERS")

        egg_filename = os.path.join(self.base_dir, "foo-1.0.egg")
        _create_egg_with_symlink(egg_filename, "foo")
        # When
        installer = EggInst(egg_filename, prefix=self.prefix)
        installer.install()

        # Then
        self.assertTrue(os.path.exists(header))
        self.assertTrue(os.path.exists(link))
        self.assertTrue(os.path.islink(link))
        self.assertEqual(os.readlink(link), "include")
        self.assertTrue(os.path.exists(os.path.join(link, "foo.h")))

    @unittest.skipIf(not SUPPORT_SYMLINK or sys.platform == "win32",
                     "this platform does not support symlink")
    def test_softlink_with_broken_entry(self):
        # Given
        path = VTK_EGG_DEFERRED_SOFTLINK

        # When
        installer = EggInst(path, prefix=self.prefix)
        installer.install()

    def test_enstaller_no_placeholder_hack(self):
        # Given
        path = _create_dummy_enstaller_egg(self.base_dir)

        # When
        with mock.patch("egginst.main.object_code.apply_placeholder_hack") \
                as m:
            installer = EggInst(path, prefix=self.prefix)
            installer.install()

        # Then
        self.assertFalse(m.called)


class TestEggInstMain(unittest.TestCase):
    def test_print_version(self):
        # XXX: this is lousy test: we'd like to at least ensure we're printing
        # the correct version, but capturing the stdout is a bit tricky. Once
        # we replace print by proper logging, we should be able to do better.
        main(["--version"])

    def test_list(self):
        # XXX: this is lousy test: we'd like to at least ensure we're printing
        # the correct packages, but capturing the stdout is a bit tricky. Once
        # we replace print by proper logging, we should be able to do better.
        main(["--list"])

    def test_install_simple(self):
        with mkdtemp() as d:
            main([DUMMY_EGG, "--prefix={0}".format(d)])

            self.assertTrue(os.path.basename(DUMMY_EGG) in list(get_installed(d)))

            main(["-r", DUMMY_EGG, "--prefix={0}".format(d)])

            self.assertFalse(os.path.basename(DUMMY_EGG) in list(get_installed(d)))

    def test_get_installed(self):
        r_installed_eggs = sorted([
            os.path.basename(DUMMY_EGG),
            os.path.basename(DUMMY_EGG_WITH_ENTRY_POINTS),
        ])

        with mkdtemp() as d:
            egginst = EggInst(DUMMY_EGG, d)
            egginst.install()

            egginst = EggInst(DUMMY_EGG_WITH_ENTRY_POINTS, d)
            egginst.install()

            installed_eggs = list(get_installed(d))
            self.assertEqual(installed_eggs, r_installed_eggs)


class Test_EggInstRemove(unittest.TestCase):
    def setUp(self):
        self.prefix = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.prefix)

    def test_simple_(self):
        # Given
        remover = EggInst(DUMMY_EGG, self.prefix)

        # When
        with testfixtures.LogCapture() as logger:
            remover.remove()

        # Then
        logger.check(
            ('egginst.main', 'ERROR',
             "Error: Can't find meta data for: 'dummy'")
        )


class TestEggInstInstall(unittest.TestCase):
    def setUp(self):
        self.base_dir = tempfile.mkdtemp()

        if sys.platform == "win32":
            self.bindir = os.path.join(self.base_dir, "Scripts")
            self.site_packages = os.path.join(self.base_dir, "lib", "site-packages")
        else:
            self.bindir = os.path.join(self.base_dir, "bin")
            self.site_packages = os.path.join(self.base_dir, "lib", "python" + PYTHON_VERSION, "site-packages")

        self.meta_dir = os.path.join(self.base_dir, "EGG-INFO")

    def tearDown(self):
        shutil.rmtree(self.base_dir)

    def test_simple(self):
        egginst = EggInst(DUMMY_EGG, self.base_dir)

        egginst.install()
        self.assertTrue(os.path.exists(os.path.join(self.site_packages, "dummy.py")))

        egginst.remove()
        self.assertFalse(os.path.exists(os.path.join(self.site_packages, "dummy.py")))

    def test_non_existing_removal(self):
        """
        Regression test for #208
        """
        # Given
        non_existing_package = "nono_le_petit_robot"

        # When/Then
        main(["--remove", non_existing_package])

    def test_entry_points(self):
        """
        Test we install console entry points correctly.
        """
        py_script = os.path.join(self.site_packages, "dummy.py")
        if sys.platform == "win32":
            wrapper_script = os.path.join(self.bindir, "dummy.exe")
        else:
            wrapper_script = os.path.join(self.bindir, "dummy")

        egginst = EggInst(DUMMY_EGG_WITH_ENTRY_POINTS, self.base_dir)

        egginst.install()
        self.assertTrue(os.path.exists(py_script))
        self.assertTrue(os.path.exists(wrapper_script))

        egginst.remove()
        self.assertFalse(os.path.exists(py_script))
        self.assertFalse(os.path.exists(wrapper_script))

    def test_appinst(self):
        """
        Test we install appinst bits correctly.
        """
        egg_path = DUMMY_EGG_WITH_APPINST
        appinst_path = os.path.join(self.meta_dir, "dummy_with_appinst", eggmeta.APPINST_PATH)

        egginst = EggInst(egg_path, self.base_dir)

        with mock.patch("appinst.install_from_dat", autospec=True) as m:
            egginst.install()
            m.assert_called_with(appinst_path, self.base_dir)

        with mock.patch("appinst.uninstall_from_dat", autospec=True) as m:
            egginst.remove()
            m.assert_called_with(appinst_path, self.base_dir)

    def test_old_appinst(self):
        """
        Test that we still work with old (<= 2.1.1) appinst, where
        [un]install_from_dat only takes one argument (no prefix).
        """
        egg_path = DUMMY_EGG_WITH_APPINST
        appinst_path = os.path.join(self.meta_dir, "dummy_with_appinst", eggmeta.APPINST_PATH)

        egginst = EggInst(egg_path, self.base_dir)

        def mocked_old_install_from_dat(x):
            pass

        def mocked_old_uninstall_from_dat(x):
            pass

        # XXX: we use autospec to enforce function taking exactly one argument,
        # otherwise the proper TypeError exception is not raised when calling
        # it with two arguments, which is how old vs new appinst is detected.
        with mock.patch("appinst.install_from_dat", autospec=mocked_old_install_from_dat) as m:
            egginst.install()
            m.assert_called_with(appinst_path)

        with mock.patch("appinst.uninstall_from_dat", autospec=mocked_old_uninstall_from_dat) as m:
            egginst.remove()
            m.assert_called_with(appinst_path)

    def test_without_appinst(self):
        """
        Test egginst does not crash when appinst is not available and we try
        installing eggs using appinst
        """
        egg_path = DUMMY_EGG_WITH_APPINST

        egginst = EggInst(egg_path, self.base_dir)

        with mock.patch("egginst.main.appinst", None):
            egginst.install()
            egginst.remove()

    def test_appinst_failed(self):
        """
        Test egginst does not crash when appinst is not available and we try
        installing eggs using appinst
        """
        egg_path = DUMMY_EGG_WITH_APPINST

        egginst = EggInst(egg_path, self.base_dir)

        with mock.patch("egginst.main.appinst.install_from_dat", side_effect=ValueError):
            egginst.install()

        with mock.patch("egginst.main.appinst.uninstall_from_dat", side_effect=ValueError):
            egginst.remove()


class TestEggInfoInstall(unittest.TestCase):
    def setUp(self):
        self.base_dir = tempfile.mkdtemp()

        if sys.platform == "win32":
            self.bindir = os.path.join(self.base_dir, "Scripts")
            self.site_packages = os.path.join(self.base_dir, "lib", "site-packages")
        else:
            self.bindir = os.path.join(self.base_dir, "bin")
            self.site_packages = os.path.join(self.base_dir, "lib", "python" + PYTHON_VERSION, "site-packages")

        self.meta_dir = os.path.join(self.base_dir, "EGG-INFO")

    def tearDown(self):
        shutil.rmtree(self.base_dir)

    def test_is_custom_egg(self):
        r_output = [
            (STANDARD_EGG, False),
            (DUMMY_EGG_WITH_APPINST, True),
            (DUMMY_EGG, True),
        ]

        for egg, expected in r_output:
            self.assertEqual(eggmeta.is_custom_egg(egg), expected)

    def test_standard_egg(self):
        custom_egg_info_base = os.path.join(self.base_dir, "EGG-INFO", "jinja2")
        egg_info_base = os.path.join(self.site_packages,
                                     "Jinja2-2.6-py2.7.egg-info")

        egg = STANDARD_EGG

        egginst = EggInst(egg, self.base_dir)
        egginst.install()

        # Check for files installed in $prefix/EGG-INFO
        for f in STANDARD_EGG_METADATA_FILES:
            path = os.path.join(custom_egg_info_base, f)
            self.assertTrue(os.path.exists(path))

        # Check for files installed in $site-packages/$package_egg_info-INFO
        for f in STANDARD_EGG_METADATA_FILES:
            path = os.path.join(egg_info_base, f)
            self.assertTrue(os.path.exists(path))

    def test_standard_egg_remove(self):
        egg = STANDARD_EGG

        with assert_same_fs(self, self.base_dir):
            egginst = EggInst(egg, self.base_dir)
            egginst.install()

            egginst.remove()

    def test_simple_custom_egg(self):
        custom_egg_info_base = os.path.join(self.base_dir, "EGG-INFO", "dummy")
        egg_info_base = os.path.join(self.site_packages,
                                     "dummy-{0}.egg-info".
                                     format("1.0.1-1"))
        egg = DUMMY_EGG

        egginst = EggInst(egg, self.base_dir)
        egginst.install()

        # Check for files installed in $prefix/EGG-INFO
        for f in DUMMY_EGG_METADATA_FILES:
            path = os.path.join(custom_egg_info_base, f)
            self.assertTrue(os.path.exists(path))

        # Check for files installed in $site-packages/$package_egg_info-INFO
        path = os.path.join(egg_info_base, "PKG-INFO")
        self.assertTrue(os.path.exists(path))

        path = os.path.join(egg_info_base, "spec/depend")
        self.assertFalse(os.path.exists(path))

        path = os.path.join(egg_info_base, "spec/summary")
        self.assertFalse(os.path.exists(path))

    def test_simple_custom_egg_remove(self):
        egg = DUMMY_EGG

        with assert_same_fs(self, self.base_dir):
            egginst = EggInst(egg, self.base_dir)
            egginst.install()
            egginst.remove()

    def test_custom_egg_with_usr_files(self):
        custom_egg_info_base = os.path.join(self.base_dir, "EGG-INFO", "nose")
        egg_info_base = os.path.join(self.site_packages,
                                     "nose-{0}.egg-info".
                                     format("1.3.0-1"))
        egg = NOSE_1_3_0

        egginst = EggInst(egg, self.base_dir)
        egginst.install()

        # Check for files installed in $prefix/EGG-INFO
        for f in DUMMY_EGG_METADATA_FILES:
            path = os.path.join(custom_egg_info_base, f)
            self.assertTrue(os.path.exists(path))

        # Check for files installed in $site-packages/$package_egg_info-INFO
        path = os.path.join(egg_info_base, "PKG-INFO")
        self.assertTrue(os.path.exists(path))

        path = os.path.join(egg_info_base, "spec/depend")
        self.assertFalse(os.path.exists(path))

        path = os.path.join(egg_info_base, "spec/summary")
        self.assertFalse(os.path.exists(path))

        path = os.path.join(egg_info_base, "usr/share/man/man1/nosetests.1")
        self.assertFalse(os.path.exists(path))

    def test_custom_egg_with_usr_files_remove(self):
        egg = NOSE_1_3_0

        with assert_same_fs(self, self.base_dir):
            egginst = EggInst(egg, self.base_dir)
            egginst.install()
            egginst.remove()

    def test_custom_egg_legacy_egg_info(self):
        custom_egg_info_base = os.path.join(self.base_dir, "EGG-INFO", "flake8")
        egg_info_base = os.path.join(self.site_packages,
                                     "flake8-2.0.0-2.egg-info")
        legacy_egg_info_base = os.path.join(self.site_packages, "flake8.egg-info")

        custom_metadata = ("PKG-INFO.bak", "requires.txt", "spec/depend",
                           "spec/summary")

        egg = LEGACY_EGG_INFO_EGG

        egginst = EggInst(egg, self.base_dir)
        egginst.install()

        # Check for files installed in $prefix/EGG-INFO
        for f in custom_metadata:
            path = os.path.join(custom_egg_info_base, f)
            self.assertTrue(os.path.exists(path))

        # Check for files installed in $site-packages/$package_egg_info-INFO
        for f in LEGACY_EGG_INFO_EGG_METADATA_FILES:
            path = os.path.join(egg_info_base, f)
            self.assertTrue(os.path.exists(path))

        for f in LEGACY_EGG_INFO_EGG_METADATA_FILES:
            path = os.path.join(legacy_egg_info_base, f)
            self.assertFalse(os.path.exists(path))

    def test_custom_egg_legacy_egg_info_remove(self):
        egg = LEGACY_EGG_INFO_EGG

        with assert_same_fs(self, self.base_dir):
            egginst = EggInst(egg, self.base_dir)
            egginst.install()
            egginst.remove()


class TestMisc(unittest.TestCase):
    def test_should_copy_custom_egg(self):
        # Given
        is_custom_egg = True

        # Then
        self.assertFalse(should_copy_in_egg_info("EGG-INFO/spec/depend",
                                                 is_custom_egg))
        self.assertTrue(should_copy_in_egg_info("EGG-INFO/SOURCES.txt",
                                                is_custom_egg))

    def test_should_copy_standard_egg(self):
        # Given
        is_custom_egg = False

        # Then
        self.assertTrue(should_copy_in_egg_info("EGG-INFO/spec/depend",
                                                is_custom_egg))
        self.assertTrue(should_copy_in_egg_info("EGG-INFO/SOURCES.txt",
                                                is_custom_egg))

    def test_is_in_legacy_info_custom_eggs(self):
        # Given
        is_custom_egg = True

        # Then
        self.assertFalse(is_in_legacy_egg_info("EGG-INFO/spec/depend",
                                               is_custom_egg))
        self.assertFalse(is_in_legacy_egg_info("dummy/__init__.py",
                                               is_custom_egg))
        self.assertTrue(is_in_legacy_egg_info("dummy.egg-info/__init__.py",
                                              is_custom_egg))
        self.assertTrue(is_in_legacy_egg_info("dummy.egg-info", is_custom_egg))
        self.assertTrue(is_in_legacy_egg_info("dummy-1.0.0.egg-info",
                                              is_custom_egg))

    def test_is_in_legacy_info_standard_eggs(self):
        # Given
        is_custom_egg = False

        # Then
        self.assertFalse(is_in_legacy_egg_info("EGG-INFO/spec/depend",
                                               is_custom_egg))
        self.assertFalse(is_in_legacy_egg_info("dummy/__init__.py",
                                               is_custom_egg))
        self.assertFalse(is_in_legacy_egg_info("dummy.egg-info/__init__.py",
                                               is_custom_egg))
        self.assertFalse(is_in_legacy_egg_info("dummy.egg-info", is_custom_egg))
        self.assertFalse(is_in_legacy_egg_info("dummy-1.0.0.egg-info",
                                               is_custom_egg))



import errno
import hashlib
import os.path
import stat

from egginst.utils import compute_md5, ensure_dir
from egginst.tests.common import (NOSE_1_3_0, SUPPORT_SYMLINK,
                                  VTK_EGG_DEFERRED_SOFTLINK, ZIP_WITH_SOFTLINK,
                                  mkdtemp)
from egginst.vendor.six import BytesIO
from egginst.vendor.six.moves import unittest

from egginst._compat import assertCountEqual
from egginst._zipfile import ZipFile


def list_files(top):
    paths = []
    for root, dirs, files in os.walk(top):
        for f in files:
            paths.append(os.path.join(os.path.relpath(root, top), f))
    return paths


def create_broken_symlink(link):
    ensure_dir(link)
    d = os.path.dirname(link)
    os.symlink(os.path.join(d, "nono_le_petit_robot"), link)


class TestZipFile(unittest.TestCase):
    def test_simple(self):
        # Given
        path = NOSE_1_3_0
        r_paths = [
            os.path.join("EGG-INFO", "entry_points.txt"),
            os.path.join("EGG-INFO", "PKG-INFO"),
            os.path.join("EGG-INFO", "spec", "depend"),
            os.path.join("EGG-INFO", "spec", "summary"),
            os.path.join("EGG-INFO", "usr", "share", "man", "man1", "nosetests.1"),
        ]

        # When
        with mkdtemp() as d:
            with ZipFile(path) as zp:
                zp.extractall(d)
            paths = list_files(d)

        # Then
        assertCountEqual(self, paths, r_paths)

    def test_extract(self):
        # Given
        path = NOSE_1_3_0
        arcname = "EGG-INFO/PKG-INFO"

        # When
        with mkdtemp() as d:
            with ZipFile(path) as zp:
                zp.extract(arcname, d)
            self.assertTrue(os.path.exists(os.path.join(d, arcname)))

    def test_extract_to(self):
        # Given
        path = NOSE_1_3_0
        arcname = "EGG-INFO/PKG-INFO"

        # When
        with mkdtemp() as d:
            with ZipFile(path) as zp:
                zp.extract_to(arcname, "FOO", d)
                extracted_data = zp.read(arcname)
            self.assertTrue(os.path.exists(os.path.join(d, "FOO")))
            self.assertEqual(compute_md5(os.path.join(d, "FOO")),
                             compute_md5(BytesIO(extracted_data)))
            self.assertFalse(os.path.exists(os.path.join(d, "EGG-INFO",
                                                         "PKG-INFO")))

    @unittest.skipIf(not SUPPORT_SYMLINK,
                     "this platform does not support symlink")
    def test_softlink(self):
        # Given
        path = ZIP_WITH_SOFTLINK

        # When/Then
        with mkdtemp() as d:
            with ZipFile(path) as zp:
                zp.extractall(d)
            paths = list_files(d)

            assertCountEqual(self, paths, [os.path.join("lib", "foo.so.1.3"),
                                           os.path.join("lib", "foo.so")])
            self.assertTrue(os.path.islink(os.path.join(d, "lib", "foo.so")))

    @unittest.skipIf(not SUPPORT_SYMLINK,
                     "this platform does not support symlink")
    def test_softlink_with_broken_entry(self):
        self.maxDiff = None

        # Given
        path = VTK_EGG_DEFERRED_SOFTLINK
        expected_files = [
            os.path.join('EGG-INFO', 'PKG-INFO'),
            os.path.join('EGG-INFO', 'inst', 'targets.dat'),
            os.path.join('EGG-INFO', 'inst', 'files_to_install.txt'),
            os.path.join('EGG-INFO', 'usr', 'lib', 'vtk-5.10', 'libvtkViews.so.5.10.1'),
            os.path.join('EGG-INFO', 'usr', 'lib', 'vtk-5.10', 'libvtkViews.so.5.10'),
            os.path.join('EGG-INFO', 'usr', 'lib', 'vtk-5.10', 'libvtkViews.so'),
            os.path.join('EGG-INFO', 'spec', 'lib-provide'),
            os.path.join('EGG-INFO', 'spec', 'depend'),
            os.path.join('EGG-INFO', 'spec', 'lib-depend'),
            os.path.join('EGG-INFO', 'spec', 'summary'),
        ]

        with mkdtemp() as d:
            existing_link = os.path.join(d, 'EGG-INFO/usr/lib/vtk-5.10/libvtkViews.so')
            create_broken_symlink(existing_link)

            # When
            with ZipFile(path) as zp:
                zp.extractall(d)
            files = list_files(d)

            # Then
            assertCountEqual(self, files, expected_files)
            path = os.path.join(d, "EGG-INFO/usr/lib/vtk-5.10/libvtkViews.so")
            self.assertTrue(os.path.islink(path))

    @unittest.skipIf(not SUPPORT_SYMLINK,
                     "this platform does not support symlink")
    def test_existing_symlink_replacement(self):
        # Ensure that when we overwrite an existing file with extract* methods,
        # we don't fail in the case a file already exists but is a symlink to a
        # file we don't have access to.
        self.maxDiff = None

        # Given
        some_data = b"some data"
        path = ZIP_WITH_SOFTLINK
        r_files = [
            os.path.join("lib", "foo.so.1.3"), os.path.join("lib", "foo.so")
        ]
        r_link = os.path.join("lib", "foo.so")
        r_file = os.path.join("lib", "foo.so.1.3")

        with mkdtemp() as tempdir:
            with mkdtemp() as tempdir2:
                def _create_read_only_file(read_only_file):
                    with open(read_only_file, "wb") as fp:
                        fp.write(some_data)

                    mode = os.stat(read_only_file)[stat.ST_MODE]
                    os.chmod(
                        read_only_file,
                        mode & ~stat.S_IWUSR & ~stat.S_IWGRP & ~stat.S_IWOTH
                    )

                    try:
                        with open(read_only_file, "wb") as fp:
                            pass
                        raise RuntimeError("Creation of RO file failed")
                    except IOError as e:
                        if e.errno != errno.EACCES:
                            raise

                def _create_link_to_ro(link_to_read_only_file):
                    # Hack: we create a symlink toward a RO file to check the
                    # destination can be overwritten
                    link_to_read_only_file = os.path.join(tempdir, r_file)
                    assert not os.path.islink(link_to_read_only_file)

                    os.unlink(link_to_read_only_file)
                    os.symlink(read_only_file, link_to_read_only_file)

                read_only_file = os.path.join(tempdir2, "read_only_file")
                _create_read_only_file(read_only_file)

                # When
                with ZipFile(path) as zp:
                    zp.extractall(tempdir)

                original_file = os.path.join(tempdir, r_file)
                original_file_md5 = compute_md5(original_file)

                _create_link_to_ro(original_file)

                with ZipFile(path) as zp:
                    zp.extractall(tempdir)
                files = list_files(tempdir)

                # Then
                assertCountEqual(self, files, r_files)
                self.assertTrue(os.path.islink(os.path.join(tempdir, r_link)))
                self.assertFalse(os.path.islink(original_file))
                self.assertEqual(compute_md5(original_file), original_file_md5)
                # Making sure we did not modify the file originally linked to
                # by the overwritten symlink
                self.assertEqual(
                    compute_md5(read_only_file), hashlib.md5(some_data).hexdigest()
                )



import os
import sys

from egginst.main import EggInst
from egginst.vendor.six.moves import unittest

from .common import mkdtemp, SUPPORT_SYMLINK

DUMMY_EGG_WITH_PROXY_SOFTLINK = os.path.join(os.path.dirname(__file__), "data",
                                             "dummy_with_proxy_softlink-1.0.0-1.egg")


class TestLinks(unittest.TestCase):
    @unittest.skipIf(not SUPPORT_SYMLINK or sys.platform == "win32",
                     "this platform does not support symlink or proxy softlink")
    def test_simple(self):
        r_link = "libfoo.so"
        r_source = "libfoo.so.0.0.0"

        with mkdtemp() as d:
            egginst = EggInst(DUMMY_EGG_WITH_PROXY_SOFTLINK, d)
            egginst.install()

            link = os.path.join(d, "lib", r_link)
            source = os.path.join(d, "lib", r_source)
            self.assertTrue(os.path.exists(link))
            self.assertTrue(os.path.exists(source))
            self.assertTrue(os.path.islink(link))
            self.assertEqual(os.readlink(link), os.path.basename(source))



from __future__ import absolute_import

import os
import platform

def _settings_root_XP():
	return os.path.join(os.environ['USERPROFILE'], 'Local Settings')

def _settings_root_Vista():
	return os.environ.get('LOCALAPPDATA', os.environ.get('ProgramData', '.'))

def _data_root_Windows():
	release, version, csd, ptype = platform.win32_ver()
	root = _settings_root_XP() if release == 'XP' else _settings_root_Vista()
	return os.path.join(root, 'Python Keyring')

def _data_root_Linux():
	"""
	Use freedesktop.org Base Dir Specfication to determine storage
	location.
	"""
	fallback = os.path.expanduser('~/.local/share')
	root = os.environ.get('XDG_DATA_HOME', None) or fallback
	return os.path.join(root, 'python_keyring')

_config_root_Windows = _data_root_Windows

def _config_root_Linux():
	"""
	Use freedesktop.org Base Dir Specfication to determine config
	location.
	"""
	fallback = os.path.expanduser('~/.local/share')
	key = 'XDG_DATA_HOME' # TODO: use XDG_CONFIG_HOME, ref #99.
	root = os.environ.get(key, None) or fallback
	return os.path.join(root, 'python_keyring')

# by default, use Unix convention
data_root = globals().get('_data_root_' + platform.system(), _data_root_Linux)
config_root = globals().get('_config_root' + platform.system(), _config_root_Linux)



import collections
import operator
import os
import os.path
import sys
import time

from egginst.eggmeta import info_from_z
from egginst.vendor.six.moves import urllib
from egginst._zipfile import ZipFile

from enstaller.errors import EnstallerException, NoSuchPackage
from enstaller.eggcollect import info_from_metadir
from enstaller.utils import compute_md5, path_to_uri, PY_VER
from enstaller.versions.pep386_workaround import PEP386WorkaroundVersion
from enstaller.versions.enpkg import EnpkgVersion


class PackageVersionInfo(object):
    def __init__(self, name, version):
        self.name = name
        self.version = version


class PackageMetadata(object):
    """
    PackageMetadataBase encompasses the metadata required to resolve
    dependencies.

    They are not attached to a repository.
    """
    @classmethod
    def from_egg(cls, path):
        """
        Create an instance from an egg filename.
        """
        with ZipFile(path) as zp:
            metadata = info_from_z(zp)
        metadata["packages"] = metadata.get("packages", [])
        return cls.from_json_dict(os.path.basename(path), metadata)

    @classmethod
    def from_json_dict(cls, key, json_dict):
        """
        Create an instance from a key (the egg filename) and metadata passed as
        a dictionary
        """
        version = EnpkgVersion.from_upstream_and_build(json_dict["version"],
                                                       json_dict["build"])
        return cls(key, json_dict["name"], version, json_dict["packages"],
                   json_dict["python"])

    def __init__(self, key, name, version, packages, python):
        self.key = key

        self.name = name
        self.version = version

        self.packages = packages
        self.python = python

    def __repr__(self):
        return "PackageMetadata('{0}-{1}', key={2!r})".format(
            self.name, self.version, self.key)

    @property
    def dependencies(self):
        # FIXME: we keep packages for backward compatibility (called as is in
        # the index).
        return self.packages

    @property
    def full_version(self):
        """
        The full version as a string (e.g. '1.8.0-1' for the numpy-1.8.0-1.egg)
        """
        return str(self.version)


class RepositoryPackageMetadata(PackageMetadata):
    """
    RepositoryPackageMetadata encompasses the full set of package metadata
    available from a repository.

    In particular, RepositoryPackageMetadata's instances know about which
    repository they are coming from through the store_location attribute.
    """
    @classmethod
    def from_egg(cls, path, store_location=""):
        """
        Create an instance from an egg filename.
        """
        with ZipFile(path) as zp:
            metadata = info_from_z(zp)

        if len(store_location) == 0:
            store_location = path_to_uri(os.path.dirname(path)) + "/"

        if not store_location.endswith("/"):
            msg = "Invalid uri for store location: {0!r} (expected an uri " \
                  "ending with '/')".format(store_location)
            raise ValueError(msg)

        metadata["packages"] = metadata.get("packages", [])
        st = os.stat(path)
        metadata["size"] = st.st_size
        metadata["md5"] = compute_md5(path)
        metadata["mtime"] = st.st_mtime
        metadata["store_location"] = store_location
        return cls.from_json_dict(os.path.basename(path), metadata)

    @classmethod
    def from_json_dict(cls, key, json_dict):
        version = EnpkgVersion.from_upstream_and_build(json_dict["version"],
                                                       json_dict["build"])
        return cls(key, json_dict["name"], version, json_dict["packages"],
                   json_dict["python"], json_dict["size"], json_dict["md5"],
                   json_dict.get("mtime", 0.0), json_dict.get("product", None),
                   json_dict.get("available", True),
                   json_dict["store_location"])

    def __init__(self, key, name, version, packages, python, size, md5,
                 mtime, product, available, store_location):
        super(RepositoryPackageMetadata, self).__init__(key, name, version,
                                                        packages, python)

        self.size = size
        self.md5 = md5

        self.mtime = mtime
        self.product = product
        self.available = available
        self.store_location = store_location

        self.type = "egg"

    @property
    def s3index_data(self):
        """
        Returns a dict that may be converted to json to re-create our legacy S3
        index content
        """
        keys = ("available", "md5", "name", "packages", "product",
                "python", "mtime", "size", "type")
        ret = dict((k, getattr(self, k)) for k in keys)
        ret["version"] = str(self.version.upstream)
        ret["build"] = self.version.build
        return ret

    @property
    def source_url(self):
        return urllib.parse.urljoin(self.store_location, self.key)

    def __repr__(self):
        template = "RepositoryPackageMetadata(" \
            "'{self.name}-{self.version}', key={self.key!r}, " \
            "available={self.available!r}, product={self.product!r}, " \
            "store_location={self.store_location!r})".format(self=self)
        return template


class InstalledPackageMetadata(PackageMetadata):
    @classmethod
    def from_egg(cls, path, ctime, store_location):
        """
        Create an instance from an egg filename.
        """
        with ZipFile(path) as zp:
            metadata = info_from_z(zp)
        metadata["packages"] = metadata.get("packages", [])
        metadata["ctime"] = ctime
        metadata["store_location"] = store_location
        metadata["key"] = os.path.basename(path)
        return cls.from_installed_meta_dict(metadata)

    @classmethod
    def from_meta_dir(cls, meta_dir):
        meta_dict = info_from_metadir(meta_dir)
        if meta_dict is None:
            message = "No installed metadata found in {0!r}".format(meta_dir)
            raise EnstallerException(message)
        else:
            return cls.from_installed_meta_dict(meta_dict)

    @classmethod
    def from_installed_meta_dict(cls, json_dict):
        key = json_dict["key"]
        name = json_dict["name"]
        upstream_version = json_dict["version"]
        build = json_dict.get("build", 1)
        version = EnpkgVersion.from_upstream_and_build(upstream_version, build)
        packages = json_dict.get("packages", [])
        python = json_dict.get("python", PY_VER)
        ctime = json_dict.get("ctime", time.ctime(0.0))
        store_location = json_dict.get("store_location", "")
        return cls(key, name, version, packages, python, ctime,
                   store_location)

    def __init__(self, key, name, version, packages, python, ctime,
                 store_location):
        super(InstalledPackageMetadata, self).__init__(key, name, version,
                                                       packages, python)

        self.ctime = ctime
        self.store_location = store_location


def egg_name_to_name_version(egg_name):
    """
    Convert a eggname (filename) to a (name, version) pair.

    Parameters
    ----------
    egg_name: str
        The egg filename

    Returns
    -------
    name: str
        The name
    version: str
        The *full* version (e.g. for 'numpy-1.8.0-1.egg', the full version is
        '1.8.0-1')
    """
    basename = os.path.splitext(os.path.basename(egg_name))[0]
    parts = basename.split("-", 1)
    if len(parts) != 2:
        raise ValueError("Invalid egg name: {0!r}".format(egg_name))
    else:
        return parts[0].lower(), parts[1]


def _valid_meta_dir_iterator(prefixes):
    for prefix in prefixes:
        egg_info_root = os.path.join(prefix, "EGG-INFO")
        if os.path.isdir(egg_info_root):
            for path in os.listdir(egg_info_root):
                meta_dir = os.path.join(egg_info_root, path)
                yield prefix, egg_info_root, meta_dir


class Repository(object):
    """
    A Repository is a set of package, and knows about which package it
    contains.
    """
    def _populate_from_prefixes(self, prefixes):
        if prefixes is None:  # pragma: nocover
            prefixes = [sys.prefix]

        for prefix, egg_info_root, meta_dir in _valid_meta_dir_iterator(prefixes):
            info = info_from_metadir(meta_dir)
            if info is not None:
                info["store_location"] = prefix

                package = \
                    InstalledPackageMetadata.from_installed_meta_dict(info)
                self.add_package(package)

    @classmethod
    def _from_prefixes(cls, prefixes=None):
        """
        Create a repository representing the *installed* packages.

        Parameters
        ----------
        prefixes: seq
            List of prefixes. [sys.prefix] by default
        """
        repository = cls()
        repository._populate_from_prefixes(prefixes)
        return repository

    def __init__(self, packages=None):
        self._name_to_packages = collections.defaultdict(list)

        self._store_info = ""

        packages = packages or []
        for package in packages:
            self.add_package(package)

    def __len__(self):
        return sum(len(self._name_to_packages[p])
                   for p in self._name_to_packages)

    def add_package(self, package_metadata):
        self._name_to_packages[package_metadata.name].append(package_metadata)

    def delete_package(self, package_metadata):
        """ Remove the given package.

        Removing a non-existent package is an error.

        Parameters
        ----------
        package_metadata : PackageMetadata
            The package to remove
        """
        if not self.has_package(package_metadata):
            msg = "Package '{0}-{1}' not found".format(
                package_metadata.name, package_metadata.version)
            raise NoSuchPackage(msg)
        else:
            candidates = [p for p in
                          self._name_to_packages[package_metadata.name]
                          if p.full_version != package_metadata.full_version]
            self._name_to_packages[package_metadata.name] = candidates

    def has_package(self, package_metadata):
        """Returns True if the given package is available in this repository

        Parameters
        ----------
        package_metadata : PackageMetadata
            The package to look for.

        Returns
        -------
        ret : bool
            True if the package is in the repository, false otherwise.
        """
        candidates = self._name_to_packages.get(package_metadata.name, [])
        for candidate in candidates:
            if candidate.full_version == package_metadata.full_version:
                return True
        return False

    def find_package(self, name, version):
        """Search for the first match of a package with the given name and
        version.

        Parameters
        ----------
        name : str
            The package name to look for.
        version : str
            The full version string to look for (e.g. '1.8.0-1').

        Returns
        -------
        package : RepositoryPackageMetadata
            The corresponding metadata.
        """
        version = EnpkgVersion.from_string(version)
        candidates = self._name_to_packages.get(name, [])
        for candidate in candidates:
            if candidate.version == version:
                return candidate
        raise NoSuchPackage("Package '{0}-{1}' not found".format(name,
                                                                 version))

    def find_package_from_requirement(self, requirement):
        """Search for latest package matching the given requirement.

        Parameters
        ----------
        requirement : Requirement
            The requirement to match for.

        Returns
        -------
        package : RepositoryPackageMetadata
            The corresponding metadata.
        """
        name = requirement.name
        version = requirement.version
        build = requirement.build
        if version is None:
            return self.find_latest_package(name)
        else:
            if build is None:
                upstream = PEP386WorkaroundVersion.from_string(version)
                candidates = [p for p in self.find_packages(name)
                              if p.version.upstream == upstream]
                candidates.sort(key=operator.attrgetter("version"))

                if len(candidates) == 0:
                    msg = "No package found for requirement {0!r}"
                    raise NoSuchPackage(msg.format(requirement))

                return candidates[-1]
            else:
                version = EnpkgVersion.from_upstream_and_build(version, build)
                return self.find_package(name, str(version))

    def find_latest_package(self, name):
        """Returns the latest package with the given name.

        Parameters
        ----------
        name : str
            The package's name

        Returns
        -------
        package : PackageMetadata
        """
        packages = self.find_sorted_packages(name)
        if len(packages) < 1:
            raise NoSuchPackage("No package with name {0!r}".format(name))
        else:
            return packages[-1]

    def find_sorted_packages(self, name):
        """Returns a list of package metadata with the given name and version,
        sorted from lowest to highest version (when possible).

        Parameters
        ----------
        name : str
            The package's name

        Returns
        -------
        packages : iterable
            Iterable of RepositoryPackageMetadata.
        """
        packages = self.find_packages(name)
        try:
            return sorted(packages,
                          key=operator.attrgetter("version"))
        except TypeError:
            # FIXME: allowing uncomparable versions should be disallowed at
            # some point
            return packages

    def find_packages(self, name, version=None):
        """ Returns a list of package metadata with the given name and version

        Parameters
        ----------
        name : str
            The package's name
        version : str or None
            If not None, the version to look for

        Returns
        -------
        packages : iterable
            Iterable of RepositoryPackageMetadata-like (order is unspecified)
        """
        candidates = self._name_to_packages.get(name, [])
        if version is None:
            return [package for package in candidates]
        else:
            return [package for package in candidates if package.full_version == version]

    def iter_packages(self):
        """Iter over each package of the repository

        Returns
        -------
        packages : iterable
            Iterable of RepositoryPackageMetadata-like.
        """
        for packages_set in self._name_to_packages.values():
            for package in packages_set:
                yield package

    def iter_most_recent_packages(self):
        """Iter over each package of the repository, but only the most recent
        version of a given package

        Returns
        -------
        packages : iterable
            Iterable of the corresponding RepositoryPackageMetadata-like
            instances.
        """
        for name, packages in self._name_to_packages.items():
            sorted_by_version = sorted(packages,
                                       key=operator.attrgetter("version"))
            yield sorted_by_version[-1]



# Code copied from colorama.win32. Original notice below
# Copyright Jonathan Hartley 2013. BSD 3-Clause license, see LICENSE file.

# from winbase.h
STDOUT = -11
STDERR = -12

try:
    import ctypes
    from ctypes import LibraryLoader
    windll = LibraryLoader(ctypes.WinDLL)
    from ctypes import wintypes
except (AttributeError, ImportError):
    windll = None
    SetConsoleTextAttribute = lambda *_: None
else:
    from ctypes import byref, Structure, POINTER

    class CONSOLE_SCREEN_BUFFER_INFO(Structure):
        """struct in wincon.h."""
        _fields_ = [
            ("dwSize", wintypes._COORD),
            ("dwCursorPosition", wintypes._COORD),
            ("wAttributes", wintypes.WORD),
            ("srWindow", wintypes.SMALL_RECT),
            ("dwMaximumWindowSize", wintypes._COORD),
        ]

        def __str__(self):
            return '(%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d)' % (
                self.dwSize.Y, self.dwSize.X,
                self.dwCursorPosition.Y, self.dwCursorPosition.X,
                self.wAttributes,
                self.srWindow.Top, self.srWindow.Left, self.srWindow.Bottom,
                self.srWindow.Right,
                self.dwMaximumWindowSize.Y, self.dwMaximumWindowSize.X
            )

    _GetStdHandle = windll.kernel32.GetStdHandle
    _GetStdHandle.argtypes = [
        wintypes.DWORD,
    ]
    _GetStdHandle.restype = wintypes.HANDLE

    _GetConsoleScreenBufferInfo = windll.kernel32.GetConsoleScreenBufferInfo
    _GetConsoleScreenBufferInfo.argtypes = [
        wintypes.HANDLE,
        POINTER(CONSOLE_SCREEN_BUFFER_INFO),
    ]
    _GetConsoleScreenBufferInfo.restype = wintypes.BOOL

    handles = {
        STDOUT: _GetStdHandle(STDOUT),
        STDERR: _GetStdHandle(STDERR),
    }

    def GetConsoleScreenBufferInfo(stream_id=STDOUT):
        handle = handles[stream_id]
        csbi = CONSOLE_SCREEN_BUFFER_INFO()
        _GetConsoleScreenBufferInfo(handle, byref(csbi))
        return csbi


def get_terminal_size():
    srWindow = GetConsoleScreenBufferInfo(STDOUT).srWindow
    columns = srWindow.Right - srWindow.Left + 1
    lines = srWindow.Bottom - srWindow.Top + 1
    return (columns, lines)



#
# (C) Copyright 2014 Enthought, Inc., Austin, TX
# All right reserved.
#
# This file is open source software distributed according to the terms in
# LICENSE.txt
#

from __future__ import absolute_import

from ._common import _PyBytes_FromStringAndSize
from . import _win32api

LOAD_LIBRARY_AS_DATAFILE = 0x2


def LoadLibraryEx(FileName, handle, flags):
    if not handle == 0:
        raise ValueError("handle != 0 not supported")
    return _win32api._LoadLibraryEx(FileName, 0, flags)


def EnumResourceTypes(hModule):
    resource_types = []

    def callback(hModule, typeid, param):
        resource_types.append(typeid)
        return True

    _win32api._EnumResourceTypes(
        hModule, _win32api.ENUMRESTYPEPROC(callback), 0)
    return resource_types


def EnumResourceNames(hModule, type_):
    resource_names = []

    def callback(hModule, type_, type_name, param):
        resource_names.append(type_name)
        return True

    _win32api._EnumResourceNames(
        hModule, type_, _win32api.ENUMRESNAMEPROC(callback), 0)
    return resource_names


def EnumResourceLanguages(hModule, type_, name):
    resource_languages = []

    def callback(hModule, type_name, res_name, language_id, param):
        resource_languages.append(language_id)
        return True

    _win32api._EnumResourceLanguages(
        hModule, type_, name, _win32api.ENUMRESLANGPROC(callback), 0)
    return resource_languages


def LoadResource(hModule, type_, name, language):
    hrsrc = _win32api._FindResourceEx(hModule, type_, name, language)
    size = _win32api._SizeofResource(hModule, hrsrc)
    hglob = _win32api._LoadResource(hModule, hrsrc)
    pointer = _win32api._LockResource(hglob)
    return _PyBytes_FromStringAndSize(pointer, size)


def FreeLibrary(hModule):
    return _win32api._FreeLibrary(hModule)



import os
import os.path
import shutil
import tempfile

from egginst.tests.common import _EGGINST_COMMON_DATA
from egginst.vendor.six.moves import unittest

from enstaller.vendor import requests, responses

from enstaller.errors import InvalidChecksum
from enstaller.fetch import _DownloadManager
from enstaller.repository import Repository, RepositoryPackageMetadata
from enstaller.utils import compute_md5

from enstaller.tests.common import mocked_session_factory


class Test_DownloadManager(unittest.TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def _create_store_and_repository(self, eggs):
        repository = Repository()
        for egg in eggs:
            path = os.path.join(_EGGINST_COMMON_DATA, egg)
            package = RepositoryPackageMetadata.from_egg(path)
            repository.add_package(package)

        return repository

    def test_fetch_simple(self):
        # Given
        filename = "nose-1.3.0-1.egg"
        repository = self._create_store_and_repository([filename])

        downloader = _DownloadManager(mocked_session_factory(self.tempdir),
                                      repository)
        downloader.fetch(filename)

        # Then
        target = os.path.join(self.tempdir, filename)
        self.assertTrue(os.path.exists(target))
        self.assertEqual(compute_md5(target),
                         repository.find_package("nose", "1.3.0-1").md5)

    def test_fetch_invalid_md5(self):
        # Given
        filename = "nose-1.3.0-1.egg"
        path = os.path.join(_EGGINST_COMMON_DATA, filename)

        repository = Repository()
        package = RepositoryPackageMetadata.from_egg(path)
        package.md5 = "a" * 32
        repository.add_package(package)

        downloader = _DownloadManager(mocked_session_factory(self.tempdir),
                                      repository)
        with self.assertRaises(InvalidChecksum):
            downloader.fetch(filename)

    def test_fetch_abort(self):
        # Given
        filename = "nose-1.3.0-1.egg"
        repository = self._create_store_and_repository([filename])

        downloader = _DownloadManager(mocked_session_factory(self.tempdir),
                                      repository)
        target = os.path.join(self.tempdir, filename)

        # When
        context = downloader.iter_fetch(filename)
        for i, chunk in enumerate(context):
            if i == 1:
                context.cancel()
                break

        # Then
        self.assertFalse(os.path.exists(target))

    def test_fetch_egg_refetch(self):
        # Given
        egg = "nose-1.3.0-1.egg"

        repository = self._create_store_and_repository([egg])

        # When
        downloader = _DownloadManager(mocked_session_factory(self.tempdir),
                                      repository)
        downloader.fetch(egg)

        # Then
        target = os.path.join(self.tempdir, egg)
        self.assertTrue(os.path.exists(target))

    def test_fetch_egg_refetch_invalid_md5(self):
        # Given
        egg = "nose-1.3.0-1.egg"
        path = os.path.join(_EGGINST_COMMON_DATA, egg)

        repository = self._create_store_and_repository([egg])

        def _corrupt_file(target):
            with open(target, "wb") as fo:
                fo.write(b"")

        # When
        downloader = _DownloadManager(mocked_session_factory(self.tempdir),
                                      repository)
        downloader.fetch(egg)

        # Then
        target = os.path.join(self.tempdir, egg)
        self.assertEqual(compute_md5(target), compute_md5(path))

        # When
        _corrupt_file(target)

        # Then
        self.assertNotEqual(compute_md5(target), compute_md5(path))

        # When
        downloader.fetch(egg, force=True)

        # Then
        self.assertEqual(compute_md5(target), compute_md5(path))

        # When/Then
        # Ensure we deal correctly with force=False when the egg is already
        # there.
        downloader.fetch(egg, force=False)

    @responses.activate
    def test_fetch_unauthorized(self):
        # Given
        filename = "nose-1.3.0-1.egg"
        url = "http://api.enthought.com/eggs/yoyo/"

        repository = Repository()

        path = os.path.join(_EGGINST_COMMON_DATA, filename)
        package = RepositoryPackageMetadata.from_egg(path, url)
        repository.add_package(package)

        responses.add(responses.GET, url + filename,
                      body='{"error": "forbidden"}',
                      status=403)
        downloader = _DownloadManager(mocked_session_factory(self.tempdir),
                                      repository)

        # When/Then
        with self.assertRaises(requests.exceptions.HTTPError):
            downloader.fetch(filename)



import os
import shutil
import tempfile
import textwrap

from egginst.vendor.six.moves import unittest

from enstaller.errors import InvalidConfiguration
from enstaller.vendor import requests

from ..auth_managers import BroodBearerTokenAuth
from .._impl import APITokenAuth, UserPasswordAuth, _encode_string_base64


FAKE_USER = "john.doe"
FAKE_PASSWORD = "fake_password"
FAKE_AUTH = UserPasswordAuth(FAKE_USER, FAKE_PASSWORD)
FAKE_CREDS = FAKE_AUTH._encoded_auth


class TestUserPasswordAuth(unittest.TestCase):
    def setUp(self):
        self.prefix = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.prefix)

    def test_from_encoded_auth(self):
        # Given
        auth_string = FAKE_CREDS
        auth_section = textwrap.dedent("""
        # A Canopy / EPD subscriber authentication is required to access the
        # Canopy / EPD repository.  To change your credentials, use the 'enpkg
        # --userpass' command, which will ask you for your email address
        # password.
        %s
        """ % ("EPD_auth = '{0}'".format(FAKE_CREDS)))

        # When
        auth = UserPasswordAuth.from_encoded_auth(auth_string)

        # Then
        self.assertEqual(auth.username, FAKE_USER)
        self.assertEqual(auth.password, FAKE_PASSWORD)
        self.assertMultiLineEqual(auth.config_string, auth_section)

    def test_invalid_from_encoded(self):
        # Given
        auth_string = _encode_string_base64("yopla")

        # When/Then
        with self.assertRaises(InvalidConfiguration):
            UserPasswordAuth.from_encoded_auth(auth_string)

    def test_empty_username(self):
        # Given
        username, password = "", ""

        # When/Then
        with self.assertRaises(InvalidConfiguration):
            UserPasswordAuth(username, password)

        # Given
        username, password = None, ""

        # When/Then
        with self.assertRaises(InvalidConfiguration):
            UserPasswordAuth(username, password)

        # Given
        username, password = "a", ""

        # When
        auth = UserPasswordAuth(username, password)

        # Then
        self.assertEqual(auth.username, "a")
        self.assertEqual(auth.password, "")

    def test_change_auth(self):
        # Given
        path = os.path.join(self.prefix, "enstaller.cfg")
        with open(path, "wt") as fp:
            fp.write("")

        auth = UserPasswordAuth("dummy", "auth")

        # When
        auth.change_auth(path)

        # Then
        with open(path) as fp:
            data = fp.read()
        self.assertEqual(data, "EPD_auth = '{0}'\n".format(auth._encoded_auth))

    def test_change_auth_existing(self):
        # Given
        auth = UserPasswordAuth("dummy", "auth")

        path = os.path.join(self.prefix, "enstaller.cfg")
        with open(path, "wt") as fp:
            fp.write("EPD_auth = '{0}'\n".format(auth._encoded_auth))

        # When
        auth = UserPasswordAuth("another_dummy", "auth")
        auth.change_auth(path)

        # Then
        with open(path) as fp:
            data = fp.read()
        self.assertEqual(data, "EPD_auth = '{0}'\n".format(auth._encoded_auth))

    def test_change_auth_with_api(self):
        # Given
        path = os.path.join(self.prefix, "enstaller.cfg")
        with open(path, "wt") as fp:
            fp.write("api_token = 'dummy_token'")

        auth = UserPasswordAuth("dummy", "auth")

        # When
        auth.change_auth(path)

        # Then
        with open(path) as fp:
            data = fp.read()
        self.assertEqual(data, "EPD_auth = '{0}'\n".format(auth._encoded_auth))

    def test_request_adapter(self):
        # Given
        auth = UserPasswordAuth("nono", "le petit robot")

        # When/Then
        self.assertIsInstance(auth.request_adapter, requests.auth.HTTPBasicAuth)
        self.assertEqual(auth.request_adapter.username, "nono")
        self.assertEqual(auth.request_adapter.password, "le petit robot")

    def test_cant_login_message(self):
        # Given
        r_message = "Could not authenticate as {0!r}".format("nono")

        # When
        auth = UserPasswordAuth("nono", "le petit robot")

        # Then
        self.assertEqual(auth.cant_login_message, r_message)


class TestAPITokenAuth(unittest.TestCase):
    def setUp(self):
        self.prefix = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.prefix)

    def test_simple(self):
        # Given
        token = "dummy auth"

        # When
        auth = APITokenAuth(token)

        # Then
        self.assertEqual(auth.config_string,
                         "api_token = '{0}'".format(token))
        self.assertEqual(auth.logged_message,
                         "logged in using API token")

    def test_change_auth(self):
        # Given
        path = os.path.join(self.prefix, "enstaller.cfg")
        with open(path, "wt") as fp:
            fp.write("")

        auth = APITokenAuth("dummy auth")

        # When
        auth.change_auth(path)

        # Then
        with open(path) as fp:
            data = fp.read()
        self.assertEqual(data, "api_token = 'dummy auth'")

    def test_change_auth_existing(self):
        # Given
        path = os.path.join(self.prefix, "enstaller.cfg")
        with open(path, "wt") as fp:
            fp.write("api_token = 'yoyo'")

        auth = APITokenAuth("dummy auth")

        # When
        auth.change_auth(path)

        # Then
        with open(path) as fp:
            data = fp.read()
        self.assertEqual(data, "api_token = 'dummy auth'")

    def test_change_auth_with_epd_auth(self):
        # Given
        path = os.path.join(self.prefix, "enstaller.cfg")
        with open(path, "wt") as fp:
            fp.write("EPD_auth = 'dummy_token'")

        auth = APITokenAuth("dummy_token")

        # When
        auth.change_auth(path)

        # Then
        with open(path) as fp:
            data = fp.read()
        self.assertEqual(data, auth.config_string)

    def test_request_adapter(self):
        # Given
        auth = APITokenAuth("nono le petit robot")

        # When/Then
        self.assertIsInstance(auth.request_adapter, BroodBearerTokenAuth)
        self.assertEqual(auth.request_adapter._token, "nono le petit robot")

    def test_cant_login_message(self):
        # Given
        r_message = ("Could not authenticate with the given token: check "
                     "your token settings")

        # When
        auth = APITokenAuth("le petit robot")

        # Then
        self.assertEqual(auth.cant_login_message, r_message)



from __future__ import print_function

import argparse
import os.path
import shutil
import sys

from egginst import logger
from egginst.main import (EggInst, get_installed, setuptools_egg_info_dir,
                          should_copy_in_egg_info)
from egginst.utils import makedirs, rm_rf


def package_egg_info_dir(egginst):
    return os.path.join(egginst.site_packages, setuptools_egg_info_dir(egginst.path))


def convert_path_to_posix(p):
    if os.path.sep != "/":
        return p.replace(os.path.sep, "/")
    else:
        return p


class EggInfoDirFixer(object):
    def __init__(self, egg, prefix=sys.prefix, dry_run=False):
        self._egginst = EggInst(egg, prefix=prefix)
        self.egg_info_dir = package_egg_info_dir(self._egginst)
        self.dry_run = dry_run

    def needs_repair(self):
        if not os.path.exists(self.egg_info_dir):
            return False
        if os.path.isdir(self.egg_info_dir):
            egg_info_dir_content = os.listdir(self.egg_info_dir)
            if "PKG-INFO" in egg_info_dir_content or "egginst.json" in egg_info_dir_content:
                return False
            return True
        elif os.path.isfile(self.egg_info_dir):
            return True

    def repair(self):
        source_egg_info_dir = self._egginst.meta_dir
        dest_egg_info_dir = package_egg_info_dir(self._egginst)

        _repair_package_dir(source_egg_info_dir, dest_egg_info_dir, self.dry_run)


def _in_place_repair(source_egg_info_dir, dest_egg_info_dir, dry_run):
    """
    Repair the given destination .egg-info directory using the given source
    egg-info directory.

    Parameters
    ----------
    source_egg_info_dir: str
        The full path to source directory (e.g.
        '...\\EGG-INFO\\<package_name>')
    dest_egg_info_dir: str
        The full path to target directory (e.g.
        '...\\site-package\\<package_name>-<version>.egg-info')
    dry_run: bool
        If True, no file is written.
    """
    makedirs(dest_egg_info_dir)
    for root, dirs, files in os.walk(source_egg_info_dir):
        for f in files:
            path = os.path.relpath(os.path.join(root, f), source_egg_info_dir)
            path = os.path.join("EGG-INFO", path)
            if should_copy_in_egg_info(convert_path_to_posix(path), True):
                source = os.path.join(root, f)
                target = os.path.join(dest_egg_info_dir,
                                      os.path.relpath(source, source_egg_info_dir))
                if dry_run:
                    logger.dry_run.info("Would copy %r to %r", source, target)
                else:
                    shutil.copy(source, target)


def _fix_pkg_info_file(egg_info_file, dest_egg_info_dir, dry_run):
    if not os.path.exists(egg_info_file):
        return
    if not os.path.isfile(egg_info_file):
        return

    source_pkg_info = egg_info_file
    target_pkg_info = os.path.join(dest_egg_info_dir, "PKG-INFO")

    if dry_run:
        logger.dry_run.info("Would copy %r to %r", source_pkg_info, target_pkg_info)
    else:
        shutil.copy(source_pkg_info, target_pkg_info)


def _repair_package_dir(source_egg_info_dir, dest_egg_info_dir, dry_run):
    if dry_run:
        _in_place_repair(source_egg_info_dir, dest_egg_info_dir, dry_run)
        _fix_pkg_info_file(dest_egg_info_dir, dest_egg_info_dir, dry_run)
        return

    working_dir = dest_egg_info_dir + ".wdir"
    temp_dir = dest_egg_info_dir + ".bak"

    # We do the transformation in a temporary directory we then move to the
    # final destination to avoid putting stalled .egg-info directories.
    makedirs(working_dir)
    try:
        _in_place_repair(source_egg_info_dir, working_dir, dry_run)
        _fix_pkg_info_file(dest_egg_info_dir, working_dir, dry_run)
        # Backup original egg-info file/dir
        os.rename(dest_egg_info_dir, temp_dir)
        # Move repaired egg-info file/dir to final destination
        os.rename(working_dir, dest_egg_info_dir)
    except BaseException:
        rm_rf(working_dir)
    else:
        rm_rf(temp_dir)


def repair(prefix, dry_run):
    """
    Repair every Enthought egg installed in the given prefix

    Parameters
    ----------
    prefix: str
        The prefix to repair
    dry_run: bool
        If True, the egg-info directory is not modified
    """
    for egg in get_installed(prefix):
        fixer = EggInfoDirFixer(egg, prefix, dry_run=dry_run)
        if fixer.needs_repair():
            fixer.repair()


def main(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    p = argparse.ArgumentParser(description="Script to repair '.egg-info' directories.")
    p.add_argument("-n", "--dry-run", help="Do not modify anything", action="store_true")
    p.add_argument("--prefix", help="The prefix to fix (default: '%(default)s')",
                   default=os.path.normpath(sys.prefix))
    ns = p.parse_args(argv)

    prefix = ns.prefix
    if not os.path.exists(prefix):
        p.error("Prefix {0} does not exist".format(prefix))

    repair(prefix, ns.dry_run)



from __future__ import print_function

from egginst._compat import PY2, input
from egginst.vendor.six.moves import urllib

import getpass
import json
import logging
import sys
import textwrap
import zlib

from os.path import abspath, expanduser, getmtime, getsize

from egginst.utils import compute_md5

from enstaller.errors import InvalidFormat
from enstaller.vendor import requests
from enstaller.versions.pep386_workaround import PEP386WorkaroundVersion
from enstaller import plat


_GZIP_MAGIC = "1f8b"

PY_VER = '%i.%i' % sys.version_info[:2]


def abs_expanduser(path):
    return abspath(expanduser(path))


def canonical(s):
    """
    return the canonical representations of a project name
    DON'T USE THIS IN NEW CODE (ONLY (STILL) HERE FOR HISTORICAL REASONS)
    """
    # eventually (once Python 2.6 repo eggs are no longer supported), this
    # function should only return s.lower()
    s = s.lower()
    s = s.replace('-', '_')
    if s == 'tables':
        s = 'pytables'
    return s


def comparable_version(version):
    """
    Given a version string (e.g. '1.3.0.dev234'), return an object which
    allows correct comparison.
    """
    return PEP386WorkaroundVersion.from_string(version)


def info_file(path):
    return dict(size=getsize(path),
                mtime=getmtime(path),
                md5=compute_md5(path))


def cleanup_url(url):
    """
    Ensure a given repo string, i.e. a string specifying a repository,
    is valid and return a cleaned up version of the string.
    """
    p = urllib.parse.urlparse(url)
    scheme, netloc, path, params, query, fragment = p[:6]

    if scheme == "":
        scheme = "file"

    if scheme == "file":
        netloc = expanduser(netloc)
        path = expanduser(path)

    if scheme in ('http', 'https', 'file'):
        if not path.endswith('/'):
            path += '/'
    else:
        raise InvalidFormat("Unsupported scheme: {0!r}".format(url))

    return urllib.parse.urlunparse((scheme, netloc, path, params, query, fragment))


def fill_url(url):
    url = url.replace('{ARCH}', plat.arch)
    url = url.replace('{SUBDIR}', plat.subdir)
    url = url.replace('{PLATFORM}', plat.custom_plat)
    return cleanup_url(url)


def path_to_uri(path):
    """Convert the given path string to a valid URI.

    It produces URI that are recognized by the windows
    shell API on windows, e.g. 'C:\\foo.txt' will be
    'file:///C:/foo.txt'"""
    return urllib.parse.urljoin("file:", urllib.request.pathname2url(path))


def uri_to_path(uri):
    """Convert a valid file uri scheme string to a native
    path.

    The returned path should be recognized by the OS and
    the native path functions, but is not guaranteed to use
    the native path separator (e.g. it could be C:/foo.txt
    on windows instead of C:\\foo.txt)."""
    urlpart = urllib.parse.urlparse(uri)
    if urlpart.scheme == "file":
        unquoted = urllib.parse.unquote(uri)
        path = unquoted[len("file://"):]
        if sys.platform == "win32" and path.startswith("/"):
            path = path[1:]
        return urllib.request.url2pathname(path)
    else:
        raise ValueError("Invalid file uri: {0}".format(uri))


def under_venv():
    # Python3 and canopy have base_prefix which is different from prefix
    # under a venv environment
    return (hasattr(sys, "real_prefix") or
            getattr(sys, "base_prefix", sys.prefix) != sys.prefix)


def real_prefix():
    if under_venv():
        if hasattr(sys, "real_prefix"):
            return sys.real_prefix
        else:
            return sys.base_prefix
    else:
        return sys.prefix


def prompt_yes_no(message, force_yes=False):
    """
    Prompt for a yes/no answer for the given message. Returns True if the
    answer is yes.

    Parameters
    ----------
    message : str
        The message to prompt the user with
    force_yes: boolean
        If True, then the message is only displayed, and the answer is assumed
        to be yes.

    """
    if force_yes:
        print(message)
        return True
    else:
        yn = input(message)
        return yn.lower() in set(['y', 'yes'])


def _bytes_to_hex(bdata):
    # Only use for tiny strings
    if PY2:
        return "".join("%02x" % (ord(c),) for c in bdata)
    else:
        return "".join("%02x" % c for c in bdata)


def decode_json_from_buffer(data):
    """
    Returns the decoded json dictionary contained in data. Optionally
    decompress the data if the buffer's data are detected as gzip-encoded.
    """
    if len(data) >= 2 and _bytes_to_hex(data[:2]) == _GZIP_MAGIC:
        # Some firewall/gateway has the "feature" of stripping Content-Encoding
        # from the response headers, without actually uncompressing the data,
        # in which case requests will give use a response object with
        # compressed data. We try to detect this case here, and decompress it
        # as requests would do if gzip format is detected.
        logging.debug("Detected compressed data with stripped header")
        try:
            data = zlib.decompress(data, 16 + zlib.MAX_WBITS)
        except (IOError, zlib.error) as e:
            # ContentDecodingError is the exception raised by requests when
            # urllib3 fails to decompress.
            raise requests.exceptions.ContentDecodingError(
                "Detected gzip-compressed response, but failed to decode it.",
                e)

    try:
        decoded_data = data.decode("utf8")
    except UnicodeDecodeError as e:
        raise ValueError("Invalid index data, try again ({0!r})".format(e))

    return json.loads(decoded_data)


def input_auth():
    """
    Prompt user for username and password.  Return (username, password)
    tuple or (None, None) if left blank.
    """
    print(textwrap.dedent("""\
        Please enter the email address and password for your Canopy / EPD
        subscription.  """))
    username = input('Email (or username): ').strip()
    if not username:
        return None, None
    return username, getpass.getpass('Password: ')



import contextlib
import sqlite3

MAX_KEY_SIZE = 256

_CREATE_TABLES = """\
CREATE TABLE IF NOT EXISTS queue
(
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    key STRING({max_key_size}) UNIQUE,
    value BLOB,
    modified_at INT NOT NULL
);
""".format(max_key_size=MAX_KEY_SIZE)

_TABLE_SIZE = "SELECT COUNT(id) FROM Queue;"

_DELETE_OLDEST = """\
DELETE
    FROM queue
    WHERE id = (
        SELECT id FROM queue
        ORDER BY modified_at ASC
        LIMIT 1
    );
"""

_DELETE_KEY = "DELETE FROM queue WHERE key = ?;"

_WRITE = """
INSERT OR REPLACE
    INTO queue (key, value, modified_at)
    VALUES( ?, ?, time('now'));
"""


class SQLiteCache(object):
    """
    A simple K/V cache with a max capacity.

    When the cache reaches its capacity, older entries are overwritten.

    Parameters
    ----------
    uri: str
        Sqlite connection string (e.g. 'foo.db', or ':memory:')
    capacity: int
        Maximum number of entries.
    """
    def __init__(self, uri, capacity=10, use_separate_connection=False):
        self._uri = uri
        self._capacity = capacity

        if use_separate_connection:
            self._cx = None
        else:
            self._cx = sqlite3.Connection(self._uri)

        self._use_separate_connection = use_separate_connection

        with self._connection_context() as cx:
            cx.execute("pragma synchronous = off;")
            cx.execute(_CREATE_TABLES)

        self._closed = False

    def close(self):
        """
        Close the cache.

        SQLite connection is closed, you cannot use any get/set/delete anymore.
        """
        self._closed = True
        if not self._use_separate_connection:
            self._cx.close()

    @contextlib.contextmanager
    def _connection_context(self):
        if self._use_separate_connection:
            cx = sqlite3.Connection(self._uri)
        else:
            cx = self._cx
        try:
            yield cx
        finally:
            if self._use_separate_connection:
                cx.close()

    def __enter__(self):
        return self

    def __exit__(self, *a, **kw):
        self.close()

    @property
    def capacity(self):
        return self._capacity

    @property
    def closed(self):
        return self._closed

    def size(self):
        with self._connection_context() as cx:
            c = cx.cursor()
            try:
                res = c.execute(_TABLE_SIZE)
                return res.fetchone()[0]
            finally:
                c.close()

    def get(self, key):
        """
        Retrieve the entry for the given key
        """
        with self._connection_context() as cx:
            c = cx.cursor()
            try:
                res = c.execute("SELECT value FROM queue WHERE key = ?;", (key,))
                row = res.fetchone()
                if row:
                    return row[0]
                else:
                    return None
            finally:
                c.close()

    def set(self, key, value):
        """
        Write the given value for the given key
        """
        with self._connection_context() as cx:
            c = cx.cursor()
            try:
                res = c.execute(_TABLE_SIZE)
                size = res.fetchone()[0]
                if size >= self.capacity:
                    c.execute(_DELETE_OLDEST)
                c.execute(_WRITE, (key, value))
                cx.commit()
            finally:
                c.close()

    def delete(self, key):
        """
        Delete the given key, value pair

        Does not raise an exception if the key does not exist.
        """
        with self._connection_context() as cx:
            c = cx.cursor()
            try:
                res = c.execute("SELECT value FROM queue WHERE key = ?;", (key,))
                row = res.fetchone()
                if row:
                    c.execute(_DELETE_KEY, (key,))
                cx.commit()
            finally:
                c.close()



import errno
import logging
import os
import re
import sys
import uuid


def win32_rename(src, dst):
    """A rename that does not fail if dst already exists.

    Note: It will still fail if dst is already opened,
    though.
    """
    try:
        os.rename(src, dst)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise
        # FIXME: this could still fail if the file is already opened. That's
        # not something we need ATM, adding unlink with POSIX semantics can be
        # done later if needed (see e.g. hg rename).
        os.unlink(dst)
        os.rename(src, dst)


def rename(src, dst):
    """Atomic rename that works on windows."""
    if sys.platform == "win32":
        return win32_rename(src, dst)
    else:
        return os.rename(src, dst)


def safe_write(target, writer, mode="wt"):
    """a 'safe' way to write to files.

    Instead of writing directly into a file, this function writes to a
    temporary file, and then rename the file to the target if no error occured.
    On most platforms, rename is atomic, so this avoids leaving stale files in
    inconsistent states.

    Parameters
    ----------
    target: str
        destination to write to
    writer: callable or data
        if callable, assumed to be function which takes one argument, a file
        descriptor, and writes content to it. Otherwise, assumed to be data
        to be directly written to target.
    mode: str
        opening mode
    """
    if not callable(writer):
        data = writer
        writer = lambda fp: fp.write(data)

    tmp_target = "%s.tmp%s" % (target, uuid.uuid4().hex)
    f = open(tmp_target, mode)
    try:
        writer(f)
    finally:
        f.close()
    rename(tmp_target, target)


def pkg_config_dir(prefix):
    """Return the full path of the pkgconfig directory for the given prefix."""
    return os.path.join(prefix, "lib", "pkgconfig")


def update_pkg_config_prefix(pc_file, prefix):
    """Overwrite the prefix variable for the given .pc pkg-config file with the
    given prefix.

    The .pc file is written in-place
    """
    if not os.path.isabs(pc_file):
        pc_file = os.path.join(pkg_config_dir(prefix), pc_file)

    pat = re.compile(r"^prefix=(.*)$", re.M)
    if os.path.exists(pc_file):
        with open(pc_file) as fp:
            data = fp.read()
            data = pat.sub("prefix={0}".format(prefix), data)
        safe_write(pc_file, data, "wt")
    else:
        # FIXME: should this be an error ?
        logging.warn("%s not found" % pc_file)



import logging
import os.path
import subprocess
import sys
import zipfile

from egginst.main import BOOTSTRAP_ARCNAME

from enstaller.fetch import _DownloadManager


logger = logging.getLogger(__name__)

# Default bootstrap code used for older versions of enstaller. Starting from
# 4.8.0, enstaller eggs (as produced by bdist_egg) will contain a
# version-specific bootstrap code for more reliability.
_DEFAULT_BOOTSTRAP_CODE = """\
import contextlib
import logging
import re
import sys


VERSION_RE = re.compile(r'''
    ^
    (?P<version>\d+\.\d+)          # minimum 'N.N'
    (?P<extraversion>(?:\.\d+)*)   # any number of extra '.N' segments
    ''', re.VERBOSE)


@contextlib.contextmanager
def disable_egginst_logging():
    logger = logging.getLogger("egginst")

    old = logger.propagate
    logger.propagate = False
    try:
        yield
    finally:
        logger.propagate = old

egg = sys.argv[1]

sys.path.insert(0, egg)
import egginst.main

# HACK: we patch argv to handle old enstallers whose main functions did not
# take an argument
with disable_egginst_logging():
    sys.argv[1:] = ["--remove", egg]
    egginst.main.main()

sys.argv[1:] = [egg]
egginst.main.main()
"""


def inplace_update(session, repository, latest):
    downloader = _DownloadManager(session, repository)
    downloader.fetch(latest.key)

    target = downloader._path(latest.key)

    with zipfile.ZipFile(target) as zp:
        try:
            bootstrap_code = zp.read(BOOTSTRAP_ARCNAME)
            logger.debug("Using bootstrap code from egg")
        except KeyError:
            logger.debug("Using default bootstrap code")
            bootstrap_code = _DEFAULT_BOOTSTRAP_CODE

    d = os.path.dirname(target)

    script_path = os.path.join(d, "bootstrap.py")
    with open(script_path, "wt") as fp:
        fp.write(bootstrap_code)

    cmd = [sys.executable, script_path, target]
    ret = subprocess.call(cmd)
    if ret != 0:
        raise ValueError("Failed to run upgrade/bootstrap code.")



from __future__ import absolute_import

import errno
import os
import shutil
import string
import zipfile

from egginst.vendor.six import text_type


ZIP_SOFTLINK_ATTRIBUTE_MAGIC = 0xA1ED0000


def is_zipinfo_symlink(zip_info):
    """Return True if the given zip_info instance refers to a symbolic link."""
    return zip_info.external_attr == ZIP_SOFTLINK_ATTRIBUTE_MAGIC


class ZipFile(zipfile.ZipFile):
    """
    A ZipFile implementation that knows how to extract soft links and allows
    overriding target destination.

    This also support context management on 2.6.
    """

    def extract_to(self, member, destination, path=None, pwd=None):
        if not isinstance(member, zipfile.ZipInfo):
            member = self.getinfo(member)

        if path is None:
            path = os.getcwd()

        return self._extract_member_to(member, destination, path, pwd)

    # Overriden so that ZipFile.extract* support softlink
    def _extract_member(self, member, targetpath, pwd):
        return self._extract_member_to(member, member.filename, targetpath, pwd)

    def _extract_symlink(self, member, link_name, pwd=None):
        source = self.read(member).decode("utf8")
        if os.path.lexists(link_name):
            os.unlink(link_name)
        os.symlink(source, link_name)
        return link_name

    # This is mostly copied from the stdlib zipfile.ZipFile._extract_member,
    # extended to allow soft link support. This needed copying to allow arcname
    # to not be based on ZipInfo.arcname
    def _extract_member_to(self, member, arcname, targetpath, pwd):
        """Extract the ZipInfo object 'member' to a physical
           file on the path targetpath.
        """
        # build the destination pathname, replacing
        # forward slashes to platform specific separators.
        arcname = arcname.replace('/', os.path.sep)

        if os.path.altsep:
            arcname = arcname.replace(os.path.altsep, os.path.sep)
        # interpret absolute pathname as relative, remove drive letter or
        # UNC path, redundant separators, "." and ".." components.
        arcname = os.path.splitdrive(arcname)[1]
        arcname = os.path.sep.join(x for x in arcname.split(os.path.sep)
                                   if x not in ('', os.path.curdir, os.path.pardir))
        if os.path.sep == '\\':
            # filter illegal characters on Windows
            illegal = ':<>|"?*'
            if isinstance(arcname, text_type):
                table = dict((ord(c), ord('_')) for c in illegal)
            else:
                table = string.maketrans(illegal, '_' * len(illegal))
            arcname = arcname.translate(table)
            # remove trailing dots
            arcname = (x.rstrip('.') for x in arcname.split(os.path.sep))
            arcname = os.path.sep.join(x for x in arcname if x)

        targetpath = os.path.join(targetpath, arcname)
        targetpath = os.path.normpath(targetpath)

        # Create all upper directories if necessary.
        upperdirs = os.path.dirname(targetpath)
        if upperdirs and not os.path.exists(upperdirs):
            os.makedirs(upperdirs)

        if member.filename[-1] == '/':
            if not os.path.isdir(targetpath):
                os.mkdir(targetpath)
            return targetpath

        if is_zipinfo_symlink(member):
            return self._extract_symlink(member, targetpath, pwd)
        else:
            source = self.open(member, pwd=pwd)
            try:
                _unlink_if_exists(targetpath)
                with open(targetpath, "wb") as target:
                    shutil.copyfileobj(source, target)
            finally:
                source.close()

            return targetpath

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


def _unlink_if_exists(p):
    try:
        os.unlink(p)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise



"""
__init__.py

Created by Kang Zhang on 2009-07-09
"""

from __future__ import absolute_import

import logging
logger = logging.getLogger('keyring')

from .core import (set_keyring, get_keyring, set_password, get_password,
                  delete_password)
from .getpassbackend import get_password as get_pass_get_password



from __future__ import absolute_import

import json

from enstaller.errors import AuthFailedError
from enstaller.vendor import requests


class UserInfo(object):
    @classmethod
    def from_json_string(cls, s):
        return cls.from_json(json.loads(s))

    @classmethod
    def from_json(cls, json_data):
        return cls(json_data["is_authenticated"],
                   json_data["first_name"],
                   json_data["last_name"],
                   json_data["has_subscription"],
                   json_data["subscription_level"])

    @classmethod
    def from_session(cls, session):
        """
        Returns the user information.

        The session must be authenticated.

        Parameters
        ----------
        session : Session
            An enstaller session, used to connect to the server to get user
            information.
        """
        # FIXME: circular import
        from .auth_managers import LegacyCanopyAuthManager

        if isinstance(session._authenticator, LegacyCanopyAuthManager):
            try:
                resp = session.get(session._authenticator.url)
            except requests.exceptions.ConnectionError as e:
                raise AuthFailedError(e)

            try:
                resp.raise_for_status()
            except requests.exceptions.HTTPError as e:
                raise AuthFailedError("Authentication error: %r" % str(e))

            return cls.from_json_string(resp.content.decode("utf8"))
        else:
            return cls(True)

    def __init__(self, is_authenticated, first_name="", last_name="",
                 has_subscription=False, subscription_level="free"):
        self.is_authenticated = is_authenticated
        self.first_name = first_name
        self.last_name = last_name
        self.has_subscription = has_subscription
        self._subscription_level = subscription_level

    @property
    def subscription_level(self):
        if self.is_authenticated and self.has_subscription:
            return 'Canopy / EPD Basic or above'
        elif self.is_authenticated and not self.has_subscription:
            return 'Canopy / EPD Free'
        else:
            return None

    def to_dict(self):
        keys = (
            "is_authenticated",
            "first_name",
            "last_name",
            "has_subscription",
            "subscription_level",
        )
        return dict((k, getattr(self, k)) for k in keys)

    def __eq__(self, other):
        return self.to_dict() == other.to_dict()


DUMMY_USER = UserInfo(False)



#
# (C) Copyright 2014 Enthought, Inc., Austin, TX
# All right reserved.
#
# This file is open source software distributed according to the terms in
# LICENSE.txt
#

from __future__ import absolute_import

import ctypes
from ctypes.wintypes import (
    BOOL, DWORD, HANDLE, HMODULE, LONG, LPCWSTR, WCHAR, WORD, HRSRC,
    HGLOBAL, LPVOID)

from ._common import LONG_PTR
from ._util import check_null, check_zero, function_factory

kernel32 = ctypes.windll.kernel32

ENUMRESTYPEPROC = ctypes.WINFUNCTYPE(BOOL, HMODULE, LONG, LONG_PTR)
ENUMRESNAMEPROC = ctypes.WINFUNCTYPE(BOOL, HMODULE, LONG, LONG, LONG_PTR)
ENUMRESLANGPROC = ctypes.WINFUNCTYPE(
    BOOL, HMODULE, WCHAR, WCHAR, WORD, LONG_PTR)

_LoadLibraryEx = function_factory(
    kernel32.LoadLibraryExW,
    [LPCWSTR, HANDLE, DWORD],
    HMODULE, check_null)

_FreeLibrary = function_factory(
    kernel32.FreeLibrary,
    [HMODULE],
    BOOL,
    check_zero)

_EnumResourceTypes = function_factory(
    kernel32.EnumResourceTypesW,
    [HMODULE, ENUMRESTYPEPROC, LONG_PTR],
    BOOL,
    check_zero)

_EnumResourceNames = function_factory(
    kernel32.EnumResourceNamesW,
    [HMODULE, DWORD, ENUMRESNAMEPROC, LONG_PTR],
    BOOL,
    check_zero)

_EnumResourceLanguages = function_factory(
    kernel32.EnumResourceLanguagesW,
    [HMODULE, DWORD, DWORD, ENUMRESLANGPROC, LONG_PTR],
    BOOL,
    check_zero)

_LoadResource = function_factory(
    kernel32.LoadResource,
    [HMODULE, HRSRC],
    HGLOBAL,
    check_zero)

_LockResource = function_factory(
    kernel32.LockResource,
    [HGLOBAL],
    LPVOID,
    check_null)

_FindResourceEx = function_factory(
    kernel32.FindResourceExW,
    [HMODULE, DWORD, DWORD, WORD],
    HRSRC,
    check_null)

_SizeofResource = function_factory(
    kernel32.SizeofResource,
    [HMODULE, HRSRC],
    DWORD,
    check_zero)



import sys
import platform


if '64' in platform.architecture()[0]:
    arch = 'amd64'
    bits = 64
else:
    arch = 'x86'
    bits = 32


def _guess_plat():
    sys_map = {'linux2': 'rh5', 'darwin': 'osx',
               'sunos5': 'sol', 'win32': 'win', 'linux': 'rh5'}
    try:
        return '%s-%d' % (sys_map[sys.platform], bits)
    except KeyError:
        return None


try:
    from custom_tools import platform as custom_plat
except ImportError:
    custom_plat = _guess_plat()


SUBDIR_MAP = {
    'win-64': 'Windows/amd64',
    'win-32': 'Windows/x86',
    'osx-64': 'MacOSX/amd64',
    'osx-32': 'MacOSX/x86',
    'rh3-64': 'RedHat/RH3_amd64',
    'rh3-32': 'RedHat/RH3_x86',
    'rh5-64': 'RedHat/RH5_amd64',
    'rh5-32': 'RedHat/RH5_x86',
    'sol-64': 'Solaris/Sol10_amd64',
    'sol-32': 'Solaris/Sol10_x86',
}

subdir = SUBDIR_MAP.get(custom_plat)
